@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2020 Red Hat, Inc.
+# Copyright (C) 2021 Red Hat, Inc.
 #
 #    This program is free software: you can redistribute it and/or modify
 #    it under the terms of the GNU General Public License as published by
@@ -21,7 +21,7 @@
     Module containing helper classes and methods which will be used by the classes and methods in
     orchestrate and execute modules
 
-    :copyright: (c) 2020 Red Hat, Inc.
+    :copyright: (c) 2021 Red Hat, Inc.
     :license: GPLv3, see LICENSE for more details.
 """
 
@@ -40,13 +40,15 @@ from shutil import copyfile
 from ansible.config.manager import ConfigManager
 from ._compat import string_types
 from .helpers import ssh_retry, exec_local_cmd_pipe, DataInjector, get_ans_verbosity, is_host_localhost, file_mgmt, \
-    gen_random_str
+    gen_random_str, check_for_var_file
 from .static.playbooks import GIT_CLONE_PLAYBOOK, SYNCHRONIZE_PLAYBOOK, \
     ADHOC_SHELL_PLAYBOOK, ADHOC_SCRIPT_PLAYBOOK
 from .exceptions import AnsibleServiceError
 from ansible.parsing.vault import VaultSecret
 from .exceptions import AnsibleVaultError
 from ._compat import RawConfigParser, VaultLib, ansible_ver, is_py2
+import configparser
+import glob
 
 LOG = getLogger(__name__)
 
@@ -177,12 +179,17 @@ class AnsibleController(object):
                         (self.ansible_inventory, playbook)
         if extra_vars is not None:
             for key in extra_vars:
-                if not isinstance(extra_vars[key], string_types):
+                if not isinstance(extra_vars[key], string_types) and key != 'file':
                     extra_var_dict = dict()
                     extra_var_dict[key] = extra_vars[key]
                     playbook_call += ' -e "%s" ' % extra_var_dict
                 elif key == "file":
-                    playbook_call += ' -e "@%s" ' % extra_vars[key]
+                    # check to see if file is a list
+                    if isinstance(extra_vars[key], list):
+                        for item in extra_vars[key]:
+                            playbook_call += ' -e "@%s" ' % item
+                    else:
+                        playbook_call += ' -e "@%s" ' % extra_vars[key]
                 else:
                     playbook_call += " -e %s=\"'%s'\"" % (key, extra_vars[key])
 
@@ -220,7 +227,7 @@ class AnsibleService(object):
 
     playbook_name = Template("cbn_execute_$type$uid.yml")
 
-    def __init__(self, config, hosts, all_hosts, ansible_options, galaxy_options=None, concurrency=None):
+    def __init__(self, config, hosts, all_hosts, ansible_options, galaxy_options=None, concurrency=None, env_var={}):
         self.hosts = hosts
         self.all_hosts = all_hosts
         self.config = config
@@ -231,7 +238,12 @@ class AnsibleService(object):
         # uuid that we can append to playbook and results file in case
         # execute/orchestrate tasks run concurrently
         self.uid = gen_random_str(5)
-        self.env_var = None
+        # setting the env variables for ansible with the os.environ
+        self.env_var = os.environ
+        if env_var:
+            # Adding any user defined env var to the os.environ
+            self.env_var = self.env_var.update(env_var)
+
         self.ans_log_path = ''
 
         # Setting ANSIBLE_LOG_PATH env variable when running actions concurrently, so each action logs gets
@@ -239,7 +251,7 @@ class AnsibleService(object):
         if concurrency == 'true':
             ans_log = 'ansible_' + self.uid + '.log'
             self.ans_log_path = os.path.join(self.config['WORKSPACE'], ans_log)
-            self.env_var = {'ANSIBLE_LOG_PATH': self.ans_log_path}
+            self.env_var.update({'ANSIBLE_LOG_PATH': self.ans_log_path})
 
         # passing the teflo's inventory directory to the Ansible Controller
         self.ans_controller = AnsibleController(os.path.abspath(self.config['INVENTORY_FOLDER']))
@@ -391,7 +403,6 @@ class AnsibleService(object):
             extra_vars.update(self.options['extra_vars'])
             # inject data into extra_vars
             extra_vars = self.injector.inject_dictionary(extra_vars)
-
         extra_vars['localhost'] = False
         if self.hosts:
             for h in self.hosts:
@@ -400,6 +411,19 @@ class AnsibleService(object):
                         extra_vars['localhost'] = True
                 elif isinstance(h, string_types) and h == 'localhost':
                     extra_vars['localhost'] = True
+        # Looking for all the variable files that teflo finds and add them as extra_vars for the ansible playbooks
+        if self.config.get('ANSIBLE_EXTRA_VARS_FILES') and \
+                self.config.get('ANSIBLE_EXTRA_VARS_FILES').lower() == 'true':
+            var_files_list = self.config.get('EXTRA_VARS_FILES')
+            file = extra_vars.get('file')
+            if file:
+                if not isinstance(file, str):
+                    extra_vars['file'].extend(var_files_list)
+                else:
+                    var_files_list.append(file)
+                    extra_vars['file'] = var_files_list
+            else:
+                extra_vars['file'] = var_files_list
 
         return extra_vars
 
@@ -539,6 +563,21 @@ class AnsibleService(object):
                     returndict["connection"] = setting.value
             return returndict
 
+    def get_playbook_path(self, coll_path=None, playbook=None):
+        """ This method is used while looking for the playbook path using fqcn
+        """
+        if coll_path and playbook:
+            fqcn_playbook_path = playbook.split('.')
+            coll_dir_path = coll_path + '/ansible_collections/' + '%s/%s' % (fqcn_playbook_path[0],
+                                                                             fqcn_playbook_path[1])
+            files = glob.glob(coll_dir_path + '/**/*.yml', recursive=True)
+            if files:
+                for path in files:
+                    if fqcn_playbook_path[-1] in path:
+                        self.logger.debug('Found Collections at: %s' % path)
+                        return path
+            return coll_dir_path
+
     def alog_update(self, folder_name=None):
         """move ansible logs to data folder/folder_name(if provided)
         :param folder_name: name of the folder under data folder to move the ansible logs
@@ -841,7 +880,7 @@ class AnsibleCredentialManager(object):
                 vaultpass = os.getenv("VAULTPASS")
             else:
                 vaultpass = self.__config.get("VAULTPASS", "")
-            if vaultpass is "":
+            if vaultpass == "":
                 raise AnsibleVaultError('No vaultpass was found, please set the \
                 vaultpass in teflo.cfg or in environment variable.')
             self.populate_credetials(self.__config, self.__config["CREDENTIAL_PATH"], vaultpass)
