@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2020 Red Hat, Inc.
+# Copyright (C) 2021 Red Hat, Inc.
 #
 #    This program is free software: you can redistribute it and/or modify
 #    it under the terms of the GNU General Public License as published by
@@ -21,7 +21,7 @@
     Module containing the core classes which teflo resources, tasks,
     providers, provisioners, etc inherit.
 
-    :copyright: (c) 2020 Red Hat, Inc.
+    :copyright: (c) 2021 Red Hat, Inc.
     :license: GPLv3, see LICENSE for more details.
 """
 import errno
@@ -35,11 +35,12 @@ from logging import config as log_config
 from time import time, sleep
 from collections import OrderedDict
 from .exceptions import TefloError, TefloResourceError, LoggerMixinError, TefloImporterError
-from .helpers import get_core_tasks_classes
+from .helpers import get_core_tasks_classes, file_mgmt
 from traceback import format_exc
 from ._compat import RawConfigParser, string_types
 from .constants import LOGGING_CONFIG
 import threading
+from .helpers import gen_random_str
 
 
 class LoggerMixin(object):
@@ -397,6 +398,21 @@ class TefloResource(LoggerMixin, TimeMixin):
 
         # every resource can have optional labels
         self._labels = list()
+        self._resource_id = gen_random_str(10)
+
+    def __eq__(self, o) -> bool:
+        return self.resource_id == o.resource_id
+
+    def __hash__(self) -> int:
+        return super().__hash__()
+
+    @property
+    def resource_id(self):
+        return self._resource_id
+
+    @resource_id.setter
+    def resource_id(self, id):
+        raise AttributeError('You cannot set resource_id after class is instantiated.')
 
     @property
     def name(self):
@@ -512,17 +528,17 @@ class TefloResource(LoggerMixin, TimeMixin):
 
         :param data: a dictionary with attributes for the resource
         """
-        for key, value in data.items():
-            # name has precedence when coming from a YAML file
-            # if name is set through name=, it will be overwritten
-            # by the YAML file properties.
-            if key == 'name':
-                self._name = value
-            elif key == 'description':
-                self._description = value
-            elif key in self._fields:
-                setattr(self, key, value)
         if data:
+            for key, value in data.items():
+                # name has precedence when coming from a YAML file
+                # if name is set through name=, it will be overwritten
+                # by the YAML file properties.
+                if key == 'name':
+                    self._name = value
+                elif key == 'description':
+                    self._description = value
+                elif key in self._fields:
+                    setattr(self, key, value)
             self.reload_tasks()
 
     def _get_task_constructors(self):
@@ -761,6 +777,7 @@ class TefloProvider(LoggerMixin, TimeMixin):
 
 class CloudProvider(TefloProvider):
     """Cloud provider class."""
+
     def __init__(self):
         super(CloudProvider, self).__init__()
         """Constructor."""
@@ -777,6 +794,7 @@ class CloudProvider(TefloProvider):
 
 class PhysicalProvider(TefloProvider):
     """Physical provider class."""
+
     def __init__(self):
         super(PhysicalProvider, self).__init__()
         """Constructor."""
@@ -785,6 +803,7 @@ class PhysicalProvider(TefloProvider):
 
 class ReportProvider(TefloProvider):
     """Report provider class."""
+
     def __init__(self):
         super(ReportProvider, self).__init__()
         """Constructor."""
@@ -1145,7 +1164,6 @@ class ProvisionerPlugin(TefloPlugin):
     """
 
     def __init__(self, asset):
-
         """Constructor.
 
         Each asset resource is linked to a provisioner that knows how to
@@ -1475,9 +1493,9 @@ class Inventory(LoggerMixin, FileLockMixin, SingletonMixin):
         self.config['INVENTORY_FOLDER'] = self.inv_dir
 
         # set the master inventory
-        self.master_inv = os.path.join(self.inv_dir, 'master-%s' % self.uid)
+        self.master_inv = os.path.join(self.inv_dir, 'inventory-%s' % self.uid)
 
-    def create_master(self, all_hosts):
+    def create_inventory(self, all_hosts):
         """Create the master ansible inventory.
         This method will create a master inventory which contains all the
         hosts in the given scenario. Each host will have a group/group:vars.
@@ -1486,12 +1504,11 @@ class Inventory(LoggerMixin, FileLockMixin, SingletonMixin):
 
             # get the lock
             self.acquire()
-
             # check for any old master inventories and delete them.
             # Check both the static inv folder as well as the .results/inventory folder for old master inventories
 
-            inv_results_dir = os.path.join(self.results_folder, 'inventory/master*')
-            stat_inv_dir = os.path.join(self.inv_dir, 'master*')
+            inv_results_dir = os.path.join(self.results_folder, 'inventory/inventory*')
+            stat_inv_dir = os.path.join(self.inv_dir, 'inventory*')
             for path in [stat_inv_dir, inv_results_dir]:
                 if glob(path):
                     for f in glob(path):
@@ -1518,10 +1535,11 @@ class Inventory(LoggerMixin, FileLockMixin, SingletonMixin):
             for host in sorted(all_hosts, key=lambda h: h.name):
                 section = host.name
                 section_vars = '%s:vars' % section
-
-                if (hasattr(host, 'role') or hasattr(host, 'groups')) and hasattr(host, 'ip_address'):
-                    for attr in ['role', 'groups']:
-                        for sect in getattr(host, attr, []):
+                # just adding the assets who have valid ip address to the inventory. This will avoid empty  entries of
+                # assets with no ip_address
+                if hasattr(host, 'ip_address') and getattr(host, 'ip_address', None):
+                    if hasattr(host, 'groups'):
+                        for sect in getattr(host, 'groups', []):
                             host_section = sect + ":children"
                             if host_section in config.sections():
                                 config.set(host_section, host.name)
@@ -1549,8 +1567,25 @@ class Inventory(LoggerMixin, FileLockMixin, SingletonMixin):
                             v = str(v)
                         config.set(section_vars, k, v)
 
-                    # write the inventory
-                    self.write_inventory(config)
+                    # get pkey var form ansible.cfg
+                    ans_path = self.config['WORKSPACE'] + '/ansible.cfg'
+                    if os.path.exists(ans_path):
+                        ans_config = RawConfigParser(allow_no_value=True)
+                        with open(ans_path) as conf:
+                            ans_config.read_file(conf)
+
+                        for k, v in ans_config['defaults'].items():
+                            config.set(section_vars, k, v)
+
+                    # get ansible group_vars file
+                    group_vars_path = self.config['WORKSPACE'] + '/ansible/group_vars/%s' % section
+                    if os.path.exists(group_vars_path):
+                        group_vars = file_mgmt('r', group_vars_path)
+                        for k, v in group_vars.items():
+                            config.set(section_vars, k, v)
+
+            # write the inventory
+            self.write_inventory(config)
 
         except Exception as ex:
             raise ex
@@ -1560,13 +1595,6 @@ class Inventory(LoggerMixin, FileLockMixin, SingletonMixin):
         self.logger.debug("Master inventory content")
         self.log_inventory_content(config)
 
-    def delete_master(self):
-        """Delete the master inventory file generated."""
-        try:
-            os.remove(self.master_inv)
-        except OSError as ex:
-            self.logger.warning(ex)
-
     def log_inventory_content(self, parser):
         # log the inventory file content
         cfg_str = ''
