@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2020 Red Hat, Inc.
+# Copyright (C) 2021 Red Hat, Inc.
 #
 #    This program is free software: you can redistribute it and/or modify
 #    it under the terms of the GNU General Public License as published by
@@ -21,12 +21,13 @@
     methods to process steps in the executes block defined within the scenario descriptor
     file. This pluging uses ansible to process these steps
 
-    :copyright: (c) 2020 Red Hat, Inc.
+    :copyright: (c) 2021 Red Hat, Inc.
     :license: GPLv3, see LICENSE for more details.
 """
 
 import ast
 import os.path
+import os
 import textwrap
 from teflo.core import ExecutorPlugin
 from teflo.exceptions import ArchiveArtifactsError, TefloExecuteError, AnsibleServiceError
@@ -67,11 +68,12 @@ class AnsibleExecutorPlugin(ExecutorPlugin):
         self.options = getattr(package, 'ansible_options', None)
         self.ignorerc = getattr(package, 'ignore_rc', False)
         self.validrc = getattr(package, 'valid_rc', None)
-
+        self.env_var = getattr(package, 'environment_vars', {})
         self.injector = DataInjector(self.all_hosts)
 
         self.ans_service = AnsibleService(self.config, self.hosts, self.all_hosts, self.options,
-                                          concurrency=self.config['TASK_CONCURRENCY']['EXECUTE'].lower())
+                                          concurrency=self.config['TASK_CONCURRENCY']['EXECUTE'].lower(),
+                                          env_var=self.env_var)
 
         self.ans_verbosity = get_ans_verbosity(self.config)
 
@@ -113,14 +115,20 @@ class AnsibleExecutorPlugin(ExecutorPlugin):
             elif validrc:
                 if result['rc'] not in validrc:
                     self.status = 1
-                    self.logger.error('Command %s failed. Host=%s rc=%d Error: %s'
-                                      % (shell['command'], result['host'], result['rc'], result['err']))
+                    self.logger.error('Command %s failed. Host=%s rc=%d\nError: Please look at the'
+                                      ' scenario log for failure.'
+                                      % (shell['command'], result['host'], result['rc']))
+                    self.logger.debug("During the failure these messages were thrown as a part of"
+                                      " stderr:\n %s" % result['err'])
 
             else:
                 if result['rc'] != 0:
                     self.status = 1
-                    self.logger.error('Command %s failed. Host=%s rc=%d Error: %s'
-                                      % (shell['command'], result['host'], result['rc'], result['err']))
+                    self.logger.error('Command %s failed. Host=%s rc=%d\nError: Please look at the'
+                                      ' scenario log for failure.'
+                                      % (shell['command'], result['host'], result['rc']))
+                    self.logger.debug("During the failure these messages were thrown as a part of"
+                                      " stderr:\n %s" % result['err'])
 
             if self.status == 1:
                 raise ArchiveArtifactsError('Command %s failed to run ' % shell['name'])
@@ -148,13 +156,19 @@ class AnsibleExecutorPlugin(ExecutorPlugin):
 
                 if result['rc'] not in validrc:
                     self.status = 1
-                    self.logger.error('Script %s failed. Host=%s rc=%d Error: %s'
-                                      % (script['name'], result['host'], result['rc'], result['err']))
+                    self.logger.error('Script %s failed. Host=%s rc=%d\nError: Please look at the'
+                                      ' scenario log for failure.'
+                                      % (script['name'], result['host'], result['rc']))
+                    self.logger.debug("During the failure these messages were thrown as a part of"
+                                      " stderr:\n %s" % result['err'])
             else:
                 if result['rc'] != 0:
                     self.status = 1
-                    self.logger.error('Script %s failed. Host=%s rc=%d Error: %s'
-                                      % (script['name'], result['host'], result['rc'], result['err']))
+                    self.logger.error('Script %s failed. Host=%s rc=%d\nError: Please look at the'
+                                      ' scenario log for failure.'
+                                      % (script['name'], result['host'], result['rc']))
+                    self.logger.debug("During the failure these messages were thrown as a part of"
+                                      " stderr:\n %s" % result['err'])
             if self.status == 1:
                 raise ArchiveArtifactsError('Script %s failed to run' % script['name'])
             else:
@@ -175,7 +189,11 @@ class AnsibleExecutorPlugin(ExecutorPlugin):
                                  % playbook['name'])
             elif results[0] != 0:
                 self.status = 1
-                raise ArchiveArtifactsError('Playbook %s failed to run' % playbook['name'])
+                if results[1]:
+                    self.logger.debug("During the failure these messages were thrown as a part of"
+                                      " stderr:\n %s " % results[1])
+                raise ArchiveArtifactsError('Playbook %s failed to run.\nPlease look at the scenario'
+                                            ' log for ansible failure.' % playbook['name'])
             else:
                 self.logger.info('Successfully executed playbook : %s' % playbook['name'])
 
@@ -210,8 +228,8 @@ class AnsibleExecutorPlugin(ExecutorPlugin):
         os.environ['ANSIBLE_LOCAL_TEMP'] = '$HOME/.ansible/tmp'
         os.environ['ANSIBLE_REMOTE_TEMP'] = '$HOME/.ansible/tmp'
 
-        # setting variable so to no display any skipped tasks
-        os.environ['DISPLAY_SKIPPED_HOSTS'] = 'False'
+        # setting variable so to not display any skipped tasks
+        os.environ['ANSIBLE_DISPLAY_SKIPPED_HOSTS'] = 'False'
 
         results = self.ans_service.run_artifact_playbook(destination, self.artifacts)
 
@@ -301,6 +319,8 @@ class AnsibleExecutorPlugin(ExecutorPlugin):
                              self.execute.testrun_results['aggregate_testrun_results']['total_tests'])
             self.logger.info(' * Failed Tests            : %s' %
                              self.execute.testrun_results['aggregate_testrun_results']['failed_tests'])
+            self.logger.info(' * Error Tests             : %s' %
+                             self.execute.testrun_results['aggregate_testrun_results']['error_tests'])
             self.logger.info(' * Skipped Tests           : %s' %
                              self.execute.testrun_results['aggregate_testrun_results']['skipped_tests'])
             self.logger.info(' * Passed Tests            : %s' %
@@ -314,6 +334,7 @@ class AnsibleExecutorPlugin(ExecutorPlugin):
                         self.logger.info(' * File Name               : %s' % keys)
                         self.logger.info(' * Total Tests             : %s' % values['total_tests'])
                         self.logger.info(' * Failed Tests            : %s' % values['failed_tests'])
+                        self.logger.info(' * Error Tests             : %s' % values['error_tests'])
                         self.logger.info(' * Skipped Tests           : %s' % values['skipped_tests'])
                         self.logger.info(' * Passed Tests            : %s' % values['passed_tests'])
                         self.logger.info('-' * 79)
