@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2020 Red Hat, Inc.
+# Copyright (C) 2021 Red Hat, Inc.
 #
 #    This program is free software: you can redistribute it and/or modify
 #    it under the terms of the GNU General Public License as published by
@@ -21,9 +21,10 @@
     Module containing classes and functions which are generic and used
     throughout the code base.
 
-    :copyright: (c) 2020 Red Hat, Inc.
+    :copyright: (c) 2021 Red Hat, Inc.
     :license: GPLv3, see LICENSE for more details.
 """
+from copy import deepcopy
 import inspect
 import json
 import os
@@ -34,6 +35,7 @@ import socket
 import string
 import subprocess
 import sys
+
 import time
 import click
 from logging import getLogger
@@ -41,23 +43,26 @@ import fnmatch
 import stat
 import jinja2
 import requests
+import ipaddress
 from paramiko import RSAKey
 from ruamel.yaml.comments import CommentedMap as OrderedDict
 from collections import OrderedDict
 from ruamel.yaml import YAML
 import yaml
-from paramiko import SSHClient, WarningPolicy
-from paramiko.ssh_exception import SSHException, BadHostKeyException, \
-    AuthenticationException
+from paramiko.ssh_exception import SSHException
 from ._compat import string_types
 from .constants import PROVISIONERS, RULE_HOST_NAMING, TASKLIST, NOTIFYSTATES
 from .exceptions import TefloError, HelpersError
 from pykwalify.core import Core
 from pykwalify.errors import CoreError, SchemaError
 from xml.etree import cElementTree as ET
-
+import socket
+from ssh.session import Session
+from ssh.key import import_privkey_file
+from ssh import options
+from ssh.exceptions import SSHError, HostKeyNotVerifiable, AuthenticationError, ConnectFailed, ConnectionLost
 import pkg_resources
-
+from ruamel.yaml import comments
 LOG = getLogger(__name__)
 
 # sentinel
@@ -217,7 +222,6 @@ def get_orchestrators_plugin_list():
 
 # Using entry point to get the executors defined in teflo's setup.py file
 def get_executors_plugin_classes():
-
     """Return all executor plugin classes discovered by teflo
     :return: The list of executor plugin classes
     """
@@ -503,8 +507,9 @@ def template_render(filepath, env_dict):
     :rtype: data stream
     """
     path, filename = os.path.split(filepath)
+
     return jinja2.Environment(loader=jinja2.FileSystemLoader(
-        path), lstrip_blocks=True, trim_blocks=True).get_template(filename).render(env_dict)
+            path), lstrip_blocks=True, trim_blocks=False).get_template(filename).render(env_dict)
 
 
 def exec_local_cmd(cmd, env_var=None):
@@ -558,10 +563,11 @@ def exec_local_cmd_pipe(cmd, logger, env_var=None):
         if output == "" and error == "" and proc.poll() is not None:
             break
         if output:
-            logger.info(output.strip())
+            logger.info(output.rstrip())
     rc = proc.poll()
     if rc != 0:
-        error = proc.stderr.readline().decode('utf-8')
+        for line in proc.stderr:
+            error = error + line.decode('utf-8')
     return rc, error
 
 
@@ -720,13 +726,9 @@ def fetch_assets(hosts, task, all_hosts=True):
             if 'all' in task[_type].hosts:
                 _hosts.append(host)
                 continue
-            if host.name in task[_type].hosts or [h for h in task[_type].hosts if h in host.name]:
+            if host.name in task[_type].hosts or [h for h in task[_type].hosts if h == host.name]:
                 _hosts.append(host)
                 continue
-            if hasattr(host, 'role'):
-                for r in host.role:
-                    if r in task[_type].hosts:
-                        _hosts.append(host)
             elif hasattr(host, 'groups'):
                 for g in host.groups:
                     if g in task[_type].hosts:
@@ -820,6 +822,11 @@ def filter_host_name(name):
     return str(result[:20]).lower()
 
 
+def is_ipv4(address):
+    ip = ipaddress.ip_address(address)
+    return isinstance(ip, ipaddress.IPv4Address)
+
+
 def ssh_retry(obj):
     """
     Decorator to check SSH Connection before method execution.
@@ -857,6 +864,7 @@ def ssh_retry(obj):
 
             sys_vars = group.vars
             server_ip = group.hosts[0].address
+            server_ip_address = socket.getaddrinfo(server_ip, None)[0][-1][0]
             LOG.info(server_ip)
 
             # skip ssh connectivity check if server is localhost
@@ -871,21 +879,31 @@ def ssh_retry(obj):
             attempt = 1
             while attempt <= MAX_ATTEMPTS:
                 try:
-                    ssh = SSHClient()
-                    ssh.set_missing_host_key_policy(WarningPolicy())
+                    # Test ssh connection
+                    pkey = import_privkey_file(server_key_file)
+
+                    if is_ipv4(server_ip_address):
+                        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+                    else:
+                        sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
+
+                    sock.connect((server_ip, server_ssh_port))
+
+                    session = Session()
+                    session.options_set(options.USER, server_user)
+                    session.options_set(options.HOST, server_ip)
+                    session.options_set_port(server_ssh_port)
+                    session.options_set(options.TIMEOUT, '5')
 
                     # Test ssh connection
-                    ssh.connect(server_ip,
-                                port=server_ssh_port,
-                                username=server_user,
-                                key_filename=server_key_file,
-                                timeout=5)
+                    session.connect()
+                    rc = session.userauth_publickey(pkey)
                     LOG.debug("Server %s - IP: %s is reachable." %
                               (group, server_ip))
-                    ssh.close()
                     break
-                except (BadHostKeyException, AuthenticationException,
-                        SSHException, socket.error) as ex:
+
+                except (SSHError, HostKeyNotVerifiable, AuthenticationError, socket.error, ConnectFailed,
+                        ConnectionLost) as ex:
                     attempt = attempt + 1
                     LOG.error(ex)
                     LOG.error("Server %s - IP: %s is unreachable." % (group,
@@ -894,6 +912,9 @@ def ssh_retry(obj):
                         LOG.info('Attempt %s of %s: retrying in %s seconds' %
                                  (attempt, MAX_ATTEMPTS, MAX_WAIT_TIME))
                         time.sleep(MAX_WAIT_TIME)
+                except Exception:
+                    LOG.error("Error occured while attempting to ssh to the host. Please verify ssh keys")
+                    return True
 
             # Check Max SSH Retries performed
             if attempt > MAX_ATTEMPTS:
@@ -932,10 +953,14 @@ def ssh_retry(obj):
 
 
 def get_ans_verbosity(config):
-    """Setting ansible verbosity
-    If the verbosity is not set in teflo.cfg, then the teflo log_level is checked.
-    If it is debug then verbotity is vvvv else it is None """
+    """Gets the Ansible verbosity level to be used by ansible or ansible-playbook commands.
 
+    Below are the options for how Ansible verbosity can be set (in order):
+        1. Ansible verbosity set within teflo.cfg
+        2. Ansible verbosity set as environment variable
+        3. Ansible verbosity is taken from teflo's log level option (debug == -vvvv)
+        4. Ansible verbosity is disabled
+    """
     if "ANSIBLE_VERBOSITY" in config and \
             config["ANSIBLE_VERBOSITY"]:
         ver = config["ANSIBLE_VERBOSITY"]
@@ -945,6 +970,17 @@ def get_ans_verbosity(config):
             LOG.warning("Ansible logging set to %s" % ans_verbosity)
         else:
             ans_verbosity = ver
+    elif os.getenv("ANSIBLE_VERBOSITY"):
+        try:
+            verbosity = os.getenv("ANSIBLE_VERBOSITY")
+            verbosity = int(verbosity)
+            if verbosity not in [0, 1, 2, 3, 4]:
+                LOG.warning(f"Ansible verbosity level: {verbosity} is invalid. Defaulting to verbosity 0.")
+                raise ValueError
+            verbosity = "v" * verbosity
+        except ValueError:
+            verbosity = None
+        ans_verbosity = verbosity
     elif config['LOG_LEVEL'] == 'debug':
         ans_verbosity = 'vvvv'
     else:
@@ -976,6 +1012,7 @@ class DataInjector(object):
     and update the string with the correct information. This makes it helpful
     when orchestrate/execute tasks require data from the hosts itself.
     """
+
     def __init__(self, hosts):
         """Constructor.
 
@@ -1288,7 +1325,238 @@ def build_artifact_regex_query(name):
     return regquery
 
 
-def validate_render_scenario(scenario, temp_data=None):
+def check_for_var_file(config, temp_data_raw=()):
+    """ This method  is for checking if variable file/directory is provided by the user in teflo.cfg under var_file key,
+     var_file.yml under the workspace , vars folder in the workspace and
+     It looks for yaml/yml files and then returns a list of file paths
+
+     Var files set every where (teflo.cfg, var_file.yml. vars folder) are collected, but the following
+     precedence is followed:
+     1. values passed over cli will override the ones set by the below options
+     2. values set in teflo.cfg will override the ones set by the below options
+     3. values set in var_file.yml will override the ones set by below options
+     4. values set in vars folder in current workspace
+
+    :param config: config object for teflo
+    :type config: config obj
+    :param temp_data_raw: cli tuple input
+    :type temp_data_raw: tuple
+    :return: var_file_list
+    :rtype: list of variable file paths
+    """
+
+    var_file_list = list()
+    var_dir = os.path.join(config.get('WORKSPACE'), 'vars')
+    workspace_var_file = os.path.join(config.get('WORKSPACE'), 'var_file.yml')
+    if config.get('VAR_FILE'):
+        default_var_file = os.path.abspath(os.path.expandvars(os.path.expanduser(config.get('VAR_FILE'))))
+    else:
+        default_var_file = ''
+
+    if os.path.exists(var_dir) and os.path.isdir(var_dir):
+        LOG.debug("Looking for .yml files as variable file under vars folder in the current workspace")
+
+        for subdir, dirs, files in os.walk(var_dir):
+            for filename in files:
+                filepath = subdir + os.sep + filename
+                if filepath.endswith(".yml"):
+                    var_file_list.append(filepath)
+
+    if os.path.exists(workspace_var_file):
+        LOG.debug("var_file.yml in current workspace added as a varaible file")
+        var_file_list.append(workspace_var_file)
+
+    if default_var_file and os.path.exists(default_var_file):
+        if os.path.isdir(default_var_file):
+            LOG.debug("Default variable file path in teflo.cfg is a directory. "
+                      "Looking for .yml files to be used as variable files")
+            for subdir, dirs, files in os.walk(default_var_file):
+                for filename in files:
+                    filepath = subdir + os.sep + filename
+                    if filepath.endswith(".yml"):
+                        var_file_list.append(filepath)
+        else:
+            LOG.debug("Default variable file found in teflo.cfg")
+            var_file_list.append(default_var_file)
+    # CHECK IF VAR_DATA WAS GIVEN BY USER IS FILE OR DIR
+    # json_val_list will get json expression input from the cli.
+    json_val_list = []
+    if temp_data_raw:
+        for item in temp_data_raw:
+            if item.endswith(".yml"):
+                var_file_list.append(item)
+            elif os.path.exists(item):
+                if os.path.isdir(item):
+                    LOG.debug("User variable file path given by CLI is a directory. "
+                              "Looking for .yml files to be used as variable files")
+                    for subdir, dirs, files in os.walk(item):
+                        for filename in files:
+                            filepath = subdir + os.sep + filename
+                            if filepath.endswith(".yml"):
+                                var_file_list.append(filepath)
+            else:
+                try:
+                    # testing if the string provided is correct json format, if its not it will throw a ValueError
+                    # Teflo will then skip that value from adding to the variable list
+                    json.loads(item)
+                except ValueError as e:
+                    LOG.info("Value entered at cli %s is Not Json Expression, will skip adding this into "
+                              "list of variables to be used" % item)
+                json_val_list.append(item)
+    # USING deepcopy() TO GET var_file_list INTO config_var_list BEFORE APPENDING JSON VALUES.
+    config_var_list = deepcopy(var_file_list)
+    config['EXTRA_VARS_FILES'] = config_var_list
+    var_file_list.extend(json_val_list)
+    return var_file_list
+
+
+def validate_brackets(input):
+    """
+    Helper funcion that validates if the input
+    is a validate string that has paired brackets
+    param input: raw variable string, something like this -> "hello {{ var }}"
+    :param type: str
+    e.x. :
+    "hello {{ var }}" is valid
+    "hello {{ var }" is invalid
+    :type config: config obj
+    :rtype: bool
+
+    """
+    left = []
+    for char in input:
+        if char == "{":
+            left.append(char)
+        if char == "}":
+            if len(left) == 0:
+                return False
+            left.pop()
+    return len(left) == 0
+
+
+def replace_brackets(input, temp_data):
+    """
+    This function replace the the refered variable by it's true value recursively
+    e.x.:
+    input_yaml:
+        a = "Hello, world!!"
+        b = "{{ hello }}" -> this is the input for this function
+    b will be conveted to "hello world" after running replace_brackerts(b, input_yaml)
+    :param input: the raw string in yaml input
+    :param type: str
+    :return new string with variables' values
+    """
+    if validate_brackets(input) and input.__contains__("{{") and input.__contains__("}}"):
+        left = []
+        key_start, key_end = 0, 0
+        replace_start, replace_end = 0, 0
+        for i in range(len(input)):
+            if input[i] == "{":
+                left.append(i)
+            elif input[i] == "}":
+                key_start, key_end = left.pop() + 1, i
+                replace_start, replace_end = left.pop(), i + 2
+                break
+
+        key = input[key_start:key_end].strip()
+        if temp_data.get(key, None) is None:
+            return input
+        if not isinstance(temp_data[key], str):
+            temp_data.update({key: preprocyaml(temp_data[key], temp_data)})
+
+        if isinstance(temp_data[key], str):
+            ret = input.replace(input[replace_start:replace_end], temp_data[key], 1)
+        elif isinstance(temp_data[key], list) or isinstance(temp_data[key], dict):
+            ret = temp_data[key]
+
+        return replace_brackets(ret, temp_data)
+    else:
+        return input
+
+
+def preprocyaml_str(input, temp_data):
+    """
+    This function is a wrapper function for replace_brackets
+    Just make it look more intuitive from its name
+    """
+    return replace_brackets(input, temp_data)
+
+
+def preprocyaml(input, temp_data):
+    """
+    This function will add the variable value to the raw string,
+    it handles all kinds of input including: string, list, dict
+    this function will add varible value to the input recursively
+    :param input: input from the yaml file
+    :param type: str, list, or dict
+    :return new value with the nested variable value added to the field in the yaml file
+    """
+    if isinstance(input, str):
+        return preprocyaml_str(input, temp_data)
+    elif isinstance(input, list):
+        new_list = []
+        for val in input:
+            new_list.append(preprocyaml(val, temp_data))
+        return new_list
+    elif isinstance(input, comments.CommentedMap):
+        new_dict = {}
+        for item in input._items():
+            if not isinstance(item[0], str):
+                result = ""
+                for key in item[0].keys():
+                    if not isinstance(temp_data[key], str):
+                        temp_data.update({key: preprocyaml(temp_data[key], temp_data)})
+                    result = result + temp_data[key]
+                return result
+            new_dict.update({item[0]: item[1]})
+        for item in new_dict.items():
+            new_dict.update({item[0]: preprocyaml(item[1], temp_data)})
+        return new_dict
+    else:
+        return input
+
+
+def preprocyaml_jinja(temp_data: dict, file_path: str = None) -> dict:
+    """
+    This function recursively render the temp_data(var_data), with itself until
+    the rendered result doesn't change any more.
+    It returns a dict version of the temp_data
+    """
+    prev_res = ""
+    res_dict = temp_data
+    result = yaml.dump(res_dict, sort_keys=False)
+
+    class NullUndefined(jinja2.Undefined):
+        def __getattr__(self, key):
+            return ''
+
+    while result != prev_res:
+        prev_res = result
+        t = jinja2.Template(result, undefined=NullUndefined)
+        result = t.render(yaml.safe_load(result))
+
+    return yaml.safe_load(result)
+
+
+def preproc_path(data_folder: str) -> str:
+    """
+    This method takes a path and remove
+    the "." and the uuid at the end of it
+    For example:
+    ./data_folder/sjf8f2vc
+    will become /data_folder
+    """
+
+    ret = ""
+    if data_folder[0] == ".":
+        ret = data_folder[1:]
+    else:
+        ret = data_folder
+    ret = ret[:ret.rindex("/")]
+    return ret
+
+
+def validate_render_scenario(scenario_path, config, temp_data_raw=()):
     """
     This method takes the absolute path of the scenario descriptor file and returns back a list of
     data streams of scenario(s) after doing the following checks:
@@ -1298,50 +1566,234 @@ def validate_render_scenario(scenario, temp_data=None):
     (4) Checks there is no yaml.safe_load error for scenario file in the include section
     :param scenario: scenario file path
     :type scenario: str
-    :param temp_data: the file path to jinja template vars data or a json dictionary of vars data
-    :type temp_data: dict or str
-    :return: scenario data stream(s)
-    :rtype: list of data streams
+    :param config: config object for teflo
+    :type config: config obj
+    :param temp_data_raw: a list of the file path to jinja template vars data or a json dictionary of vars data
+    :type temp_data_raw: dict or str
+    :return scenario_graph
+    :rtype: ScenarioGraph
     """
-    scenario_stream_list = list()
 
-    if temp_data:
-        if os.path.isfile(temp_data):
-            temp_data = file_mgmt('r', temp_data)
-        else:
-            temp_data = json.loads(temp_data)
-        # Updating th eos.environ to the variable data
-        temp_data.update(os.environ)
-    else:
-        # if temp_data isNone, sending only the os.environment
-        temp_data = os.environ
+    # Click gives us a tuple, by default
+    var_file_list = check_for_var_file(config, temp_data_raw)
+    # Convert each item to an object, then reduce them all back to one
+    temp_data_objs = [file_mgmt("r", t) if os.path.isfile(t)
+                                  else json.loads(t) for t in var_file_list]
+
+    # Reduce it down to a single object we can work with
+    temp_data = {}
+    [temp_data.update(t) for t in temp_data_objs]
+
+    for item in temp_data.items():
+        temp_data.update({item[0]: preprocyaml(item[1], temp_data)})
+        # if the processed value is not parsable by jinja2 engine,
+        # we should make it to "" then
+        if isinstance(item[1], str):
+            temp_dict = {}
+            temp_dict[item[0]] = item[1]
+            try:
+                preprocyaml_jinja(temp_dict)
+            except jinja2.exceptions.TemplateSyntaxError:
+                temp_data[item[0]] = ""
+    temp_data = preprocyaml_jinja(temp_data)
+    temp_data.update(os.environ)
+
     try:
-        data = yaml.safe_load(template_render(scenario, temp_data))
-        # adding master scenario as the first scenario data stream
-        scenario_stream_list.append(template_render(scenario, temp_data))
-        if 'include' in data.keys():
-            include_item = data['include']
-            include_template = list()
-            if include_item is not None:
-                for item in include_item:
-                    if os.path.isfile(item):
-                        item = os.path.abspath(item)
-                        # check to verify the data in included scenario is valid
-                        try:
-                            yaml.safe_load(template_render(item, temp_data))
-                            include_template.append(template_render(item, temp_data))
-                        except yaml.YAMLError:
-                            # raising Teflo error to differentiate the yaml issue is with included scenario
-                            raise TefloError('Error loading updated included scenario data!')
-                    else:
-                        # raising HelperError if included file is invalid or included section is empty
-                        raise HelpersError('Included File is invalid or Include section is empty .'
-                                           'You have to provide valid scenario files to be included.')
-                scenario_stream_list.extend(include_template)
+        # Build a scenario graph
+        scenario_graph = build_scenario_graph(root_scenario_path=scenario_path,
+                                              root_scenario_temp_data=temp_data, config=config)
+
     except yaml.YAMLError as e:
         # here raising yaml error to differentiate yaml issue is with main scenario
         raise e
-    return scenario_stream_list
+    return scenario_graph
+
+
+def build_scenario_graph(root_scenario_path: str, config, root_scenario_temp_data):
+    '''
+    This method builds a scenario graph with the root_sceanrio_path and root_scenario_temp_data
+    We utilized the iterative way to implement this, so it will not cause stackoverflow even
+    with over 1000 linked included sdfs
+
+    :param root_scenario_path: the root sdf path
+    :type root_scenario_path: str
+    :param root_scenario_temp_data: the temp data that all sdf need to render with
+    :type root_scenario_temp_data: dict
+    '''
+    # Must import these two libs here to avoid circular import for Python intepreter
+    from .resources import Scenario
+    from .utils.scenario_graph import ScenarioGraph
+    yaml.safe_load(template_render(root_scenario_path, root_scenario_temp_data))
+    root_scenario_yaml_data = template_render(root_scenario_path, root_scenario_temp_data)
+    root_scenario = Scenario(config=config, path=os.path.basename(root_scenario_path))
+    root_scenario.fullpath = root_scenario_path
+    root_scenario.yaml_data = root_scenario_yaml_data
+    # deepcopy will work only under pythonversion > 3.9
+    # leave this line here for future improvement
+    # root_config = deepcopy(config)
+
+    def addAllIncludes(parent_scenario: Scenario, checked_list: dict):
+        """
+        This method add all included child sdfs to the parent_scenario
+        as its child_scenarios
+
+        :param parent_scenario: the parent scenario object
+        :type root_scenario_path: Scenario
+        :param checked_list: the checked list for all visited Scenarios
+        :type checked_list: dict
+        """
+        if parent_scenario is None:
+            return
+
+        data = yaml.safe_load(parent_scenario.yaml_data)
+
+        # process the parent sc here, (process its remote_workspace section,include section)
+
+        def process_remote_workspace(remote_workspaces: list):
+
+            if config["REMOTE_WORKSPACE_DOWNLOAD_LOCATION"][-1] != "/":
+                config["REMOTE_WORKSPACE_DOWNLOAD_LOCATION"] = config["REMOTE_WORKSPACE_DOWNLOAD_LOCATION"] + "/"
+            ret = {}
+            for workspace in remote_workspaces:
+                url = workspace.get("workspace_url", None)
+                short_name_of_workspace = workspace.get("alias_name", None)
+                if url is None:
+                    raise TefloError(
+                            "Your format of the imported remote_workspace is incorrect ")
+                cmd = 'git clone ' + url + ' ' + config["REMOTE_WORKSPACE_DOWNLOAD_LOCATION"] + short_name_of_workspace
+                if not os.path.isdir(config["REMOTE_WORKSPACE_DOWNLOAD_LOCATION"] + short_name_of_workspace):
+                    result = exec_local_cmd(cmd)
+                    if result[0] != 0:
+                        raise TefloError("Remote remote_workspaces download failed!!")
+
+                # all remote remote_workspaces should be stored in the same place
+                # (root_workspace/.teflo_remote_workspace_cache/)
+                # so we should use root_config instead of parent_scenario.config,
+                # which always contains the root sdf workspace path
+                remote_workspace_path = os.path.join(
+                        config["WORKSPACE"], config["REMOTE_WORKSPACE_DOWNLOAD_LOCATION"] + short_name_of_workspace)
+                ret[short_name_of_workspace] = remote_workspace_path
+            return ret
+
+        workspace_info = None
+        if 'remote_workspace' in data.keys() and data['remote_workspace'] is not None \
+                and len(data['remote_workspace']) is not 0:
+            remote_workspaces = data['remote_workspace']
+            workspace_info = process_remote_workspace(remote_workspaces)
+
+        def process_path(path: str, workspace_info: dict):
+            # ex remote_workspace/sdf.yml
+            if path is None or path == "":
+                return None, None
+            if workspace_info is None:
+                return None, None
+            real_path = path.split("/")
+            if workspace_info.get(real_path[0], None) is not None:
+                real_path[0] = workspace_info.get(real_path[0], None)
+            else:
+                return path, None
+            ret = "/".join(real_path)
+            # return processed full path and the remote workspace dir name
+            return ret, real_path[0]
+
+        if 'include' in data.keys() and data['include'] is not None:
+            include_item = data['include']
+            for item in include_item:
+                if checked_list.get(item) is not None:
+                    raise TefloError(
+                        "Your scenario has an import cycle. \
+                            %s is already a node in the scenario graph, It cannot be added again. "
+                        % item)
+                # this config is from the root teflo project which means the current workspace's teflo.cfg
+                sc_fullpath = os.path.join(parent_scenario.config['WORKSPACE'], item)
+                sc_abspath = item
+                # return processed path and the remote short name
+                remote_path = process_path(item, workspace_info)
+                sc_fullpath_is_valid = os.path.isfile(sc_fullpath)
+                sc_abspath_is_valid = os.path.isfile(sc_abspath)
+                remote_path_is_valid = os.path.isfile(remote_path[0]) if \
+                    remote_path[0] else False
+
+                if sc_fullpath_is_valid or sc_abspath_is_valid or remote_path_is_valid:
+                    path = os.path.basename(item)
+                    if os.path.isfile(sc_fullpath):
+                        item = sc_fullpath
+                    elif os.path.isfile(sc_abspath):
+                        item = sc_abspath
+                    elif remote_path[0] and os.path.isfile(remote_path[0]):
+                        item = remote_path[0]
+                    # check to verify the data in included scenario is valid
+                    try:
+                        yaml.safe_load(template_render(item, root_scenario_temp_data))
+                        from .utils.config import Config
+                        root_config = Config()
+                        root_config.load()
+                        for xxx in parent_scenario.config.items():
+                            root_config[xxx[0]] = xxx[1]
+                        child_sc = Scenario(config=root_config, path=path)
+                        # change workspace if this is a remote sc
+                        if workspace_info is not None and remote_path[1]:
+                            child_sc.config["WORKSPACE"] = remote_path[1]
+                        else:
+                            child_sc.config["WORKSPACE"] = parent_scenario.config.get("WORKSPACE")
+                        child_sc.fullpath = sc_fullpath if os.path.isfile(sc_fullpath) else sc_abspath
+                        child_sc.yaml_data = template_render(item, root_scenario_temp_data)
+
+                        parent_scenario.add_child_scenario(child_sc)
+                        child_sc.my_parent = parent_scenario
+
+                        child_sc_loop = child_sc
+                        while child_sc_loop.my_parent is not None:
+                            child_sc_loop.my_parent.children_size += 1
+                            child_sc_loop = child_sc_loop.my_parent
+
+                        # use filename because the scenario name could
+                        # contain some special characters, which is not good for
+                        # file generation
+                        if preproc_path(config['RESULTS_FOLDER']) in child_sc.fullpath or \
+                                preproc_path(config["DATA_FOLDER"]) in child_sc.fullpath:
+                            parent_scenario.included_scenario_path = os.path.join(
+                                config['RESULTS_FOLDER'], child_sc.path)
+                        else:
+                            parent_scenario.included_scenario_path = os.path.join(
+                                config['RESULTS_FOLDER'], child_sc.path.split(".")[0] + "_results.yml")
+                    except yaml.YAMLError as err:
+                        # raising Teflo error to differentiate the yaml issue is with included scenario
+                        raise TefloError('Error loading included '
+                                         'scenario data! ' + item + str(err.problem_mark))
+                else:
+                    raise TefloError('Included File is invalid or Include section is empty.'
+                                     ' You have to provide valid scenario files to be included.')
+
+    def include(unchecked_list: list, checked_list: dict):
+        """
+        This method will build a scenario linked graph
+        It reads from the unckecked_list for unread scenarios
+        and add all its child to it, then add all it's children
+        to the unchecked list for next iteration
+
+        :param unchecked_list: the unchecked list of scenarios
+        :type unchecked_list: list
+        :param checked_list: the checked list for all visited Scenarios
+        :type checked_list: dict
+        """
+
+        # unchecked_list is a queue-liked list, we keep this as channel
+        # to maintain all unchecked scenarios
+        while len(unchecked_list) != 0:
+            sc: Scenario = unchecked_list.pop(0)
+            checked_list[sc.path] = sc.path
+            addAllIncludes(sc, checked_list)
+            # We need to remove the checked scenario from the unchecked_list
+            # after we addAllIncludes for it
+            if sc.child_scenarios != []:
+                for sc in sc.child_scenarios:
+                    unchecked_list.append(sc)
+
+    include([root_scenario], {})
+    scenario_graph = ScenarioGraph(root_scenario, iterate_method=config.get("INCLUDED_SDF_ITERATE_METHOD", "by_level"),
+                                   scenario_vars=root_scenario_temp_data)
+    return scenario_graph
 
 
 def ssh_key_file_generator(workspace, ssh_key_param):
@@ -1419,7 +1871,7 @@ def set_task_class_concurrency(task, resource):
     :return: TefloTask class
     """
     val = getattr(resource, 'config')['TASK_CONCURRENCY'].get(task['task'].__task_name__.upper())
-    if val == 'True':
+    if val.lower() == 'true':
         val = True
     else:
         val = False
@@ -1466,28 +1918,36 @@ def sort_tasklist(user_tasks):
         return sorted(user_tasks, key=NOTIFYSTATES.index)
 
 
-def validate_cli_scenario_option(ctx, scenario, vars_data=None):
+def validate_cli_scenario_option(ctx, scenario, config, vars_data=None):
     # Make sure the file exists and gets its absolute path
     if scenario is not None and os.path.isfile(scenario):
         scenario = os.path.abspath(scenario)
     else:
         click.echo('You have to provide a valid scenario file.')
-        ctx.exit()
+        ctx.exit(1)
 
     # Checking if include section is present and getting validated scenario stream/s
     try:
-        scenario_stream = validate_render_scenario(scenario, vars_data)
-        return scenario_stream
-    except yaml.YAMLError:
-        click.echo('Error loading updated scenario data!')
-        ctx.exit()
+        scenario_graph = validate_render_scenario(scenario, config, vars_data)
+        return scenario_graph
+    except yaml.YAMLError as err:
+        click.echo('Error loading scenario data! %s' % err)
+        ctx.exit(1)
     except HelpersError:
         click.echo('Included File is invalid or Include section is empty.'
                    'You have to provide valid scenario files to be included.')
-        ctx.exit()
-    except TefloError:
-        click.echo('Error loading updated included scenario data!')
-        ctx.exit()
+        ctx.exit(1)
+    except TefloError as err:
+        click.echo('%s' % err.message)
+        ctx.exit(1)
+    except jinja2.exceptions.UndefinedError as err:
+        curframe = inspect.currentframe()
+        calframe = inspect.getouterframes(curframe, 2)
+        if calframe[1][3] == 'show':
+            click.echo("\n\nYou need to use --vars-data to fill your variables for show command")
+            ctx.exit(1)
+        else:
+            raise TefloError("You need to fill your variable with --vars-data label")
 
 
 def create_individual_testrun_results(artifact_locations, config):
@@ -1513,6 +1973,7 @@ def create_individual_testrun_results(artifact_locations, config):
             trun = dict()
             trun['total_tests'] = 0
             trun['failed_tests'] = 0
+            trun['error_tests'] = 0
             trun['skipped_tests'] = 0
             trun['passed_tests'] = 0
             tree = ET.parse(path)
@@ -1530,10 +1991,14 @@ def create_individual_testrun_results(artifact_locations, config):
                     trun['failed_tests'] = trun['failed_tests'] + len([testcase.find('failure')
                                                                         for testcase in test.findall('testcase')
                                                                         if testcase.findall('failure')])
+                    trun['error_tests'] = trun['error_tests'] + len([testcase.find('error')
+                                                                        for testcase in test.findall('testcase')
+                                                                        if testcase.findall('error')])
                     trun['skipped_tests'] = trun['skipped_tests'] + len([testcase.find('skipped')
                                                                          for testcase in test.findall('testcase')
                                                                          if testcase.findall('skipped')])
-                    trun['passed_tests'] = trun['total_tests'] - trun['failed_tests'] - trun['skipped_tests']
+                    trun['passed_tests'] = trun['total_tests'] - trun['failed_tests'] - trun['error_tests'] -\
+                                           trun['skipped_tests']
                 individual_res.append({os.path.basename(path): trun})
             else:
                 LOG.warning("The xml file %s does not have the correct format (no 'testsuite' or 'testsuites'"
@@ -1554,6 +2019,7 @@ def create_aggregate_testrun_results(individual_results):
     """
     total_tests = 0
     failed_tests = 0
+    error_tests = 0
     skipped_tests = 0
     passed_tests = 0
     agg_results = dict()
@@ -1562,11 +2028,13 @@ def create_aggregate_testrun_results(individual_results):
         for val in run.values():
             total_tests += val['total_tests']
             failed_tests += val['failed_tests']
+            error_tests += val['error_tests']
             skipped_tests += val['skipped_tests']
             passed_tests += val['passed_tests']
 
     agg_results.update(aggregate_testrun_results=dict(total_tests=total_tests,
                                                       failed_tests=failed_tests,
+                                                      error_tests=error_tests,
                                                       skipped_tests=skipped_tests,
                                                       passed_tests=passed_tests
                                                       ))
@@ -1597,7 +2065,7 @@ def generate_default_template_vars(scenario, notification):
     """
     Default template dictionary created to be used
     when rendering the default notification template.
-    :return:
+    :return: temp_dict dict dictionary of selected variables for rendering notification templates
     """
 
     passed_tasks = getattr(scenario, 'passed_tasks', [])
@@ -1605,6 +2073,10 @@ def generate_default_template_vars(scenario, notification):
 
     temp_dict = dict(scenario=scenario)
 
+    temp_dict['scenario_graph'] = scenario.__getattribute__('scenario_graph')
+
+    temp_dict['scenario_vars'] = temp_dict.get('scenario_graph').__getattribute__('scenario_vars')
+
     if getattr(notification, 'on_start', False):
         temp_dict['passed_tasks'] = passed_tasks[-1]
         return temp_dict
@@ -1616,3 +2088,20 @@ def generate_default_template_vars(scenario, notification):
         temp_dict['failed_tasks'] = ','.join(failed_tasks)
 
     return temp_dict
+
+
+class StatusPageHelper(object):
+
+    def __init__(self, proxyserver_url):
+        self.proxyserver_url = proxyserver_url
+
+    def get_info(self):
+        ret = {}
+        try:
+            info = requests.get(self.proxyserver_url + "/components").json()['components']
+            for item in info:
+                if item.get("name") is not None:
+                    ret[item["name"]] = item
+        except Exception:
+            raise TefloError("The url is incorrect for resource check")
+        return ret
