@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2020 Red Hat, Inc.
+# Copyright (C) 2021 Red Hat, Inc.
 #
 #    This program is free software: you can redistribute it and/or modify
 #    it under the terms of the GNU General Public License as published by
@@ -21,7 +21,7 @@
     Teflo's own Beaker provisioner. This module handles everything from
     authentication to creating/deleting resources in Beaker.
 
-    :copyright: (c) 2020 Red Hat, Inc.
+    :copyright: (c) 2021 Red Hat, Inc.
     :license: GPLv3, see LICENSE for more details.
 """
 from __future__ import unicode_literals
@@ -141,7 +141,7 @@ class BeakerClientProvisionerPlugin(ProvisionerPlugin):
 
         # set attributes for beaker xml object
         for key, value in self.provider_params.items():
-            if key is not 'name':
+            if key != 'name':
                 if value:
                     setattr(self.bkr_xml, key, value)
 
@@ -151,7 +151,7 @@ class BeakerClientProvisionerPlugin(ProvisionerPlugin):
         )
 
         if 'force' in self.bkr_xml.hrname:
-            self.logger.warning('Force was specified as a host_require_option.'
+            self.logger.warning('Force was specified as a host_require_option. '
                                 'Any other host_require_options will be ignored since '
                                 'force is a mutually exclusive option in beaker.')
         # format beaker client command to run
@@ -224,7 +224,7 @@ class BeakerClientProvisionerPlugin(ProvisionerPlugin):
         self.logger.debug('Beaker timeout limit: %s.' % wait)
 
         # check Beaker status every 60 seconds
-        total_attempts = wait / 60
+        total_attempts = int(wait / 60)
 
         attempt = 0
         while wait > 0:
@@ -465,12 +465,22 @@ class BeakerClientProvisionerPlugin(ProvisionerPlugin):
         ssh_con = paramiko.SSHClient()
         ssh_con.set_missing_host_key_policy(paramiko.AutoAddPolicy())
 
+        ssh_dir = "/root/.ssh"
+
         try:
             ssh_con.connect(hostname=ip,
                             username=username,
                             password=password)
             sftp = ssh_con.open_sftp()
-            sftp.put(public_key, '/root/.ssh/authorized_keys')
+
+            try:
+                sftp.stat(ssh_dir)
+            except FileNotFoundError:
+                self.logger.warning(f"Directory {ssh_dir} does not exist on {ip}. Creating directory "
+                                    f"before sending key.")
+                sftp.mkdir(ssh_dir, mode=0o755)
+
+            sftp.put(public_key, f"{ssh_dir}/authorized_keys")
         except paramiko.SSHException as ex:
             raise BeakerProvisionerError(
                 'Failed to connect to remote system: %s' % ex
@@ -567,6 +577,7 @@ class BeakerXML(object):
         self._kernel_options_post = ""
         self._kickstart = ""
         self._ksmeta = ""
+        self._ksappends = []
         self._remove_task = False
         self._virtmachine = False
         self._virtcapable = False
@@ -675,6 +686,11 @@ class BeakerXML(object):
         if self.ksmeta != "":
             self.cmd += " --ks-meta '" + " ".join(self.ksmeta) + "'"
 
+        # Set kickstart append script
+        if self.ksappends != []:
+            for ksappend in self.ksappends:
+                self.cmd += " --ks-append '" + ksappend + "'"
+
         # Set debug and dryrun
         self.cmd += " --debug --dryrun"
 
@@ -852,6 +868,17 @@ class BeakerXML(object):
         :param meta: meta data"""
         self._ksmeta = meta
 
+    @property
+    def ksappends(self):
+        """Return the kick start append scripts."""
+        return self._ksappends
+
+    @ksappends.setter
+    def ksappends(self, scripts):
+        """Set kick start append scripts.
+        :param scripts: ksappend scripts"""
+        self._ksappends = scripts
+
     @property
     def arch(self):
         """Return the arch."""
