@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2020 Red Hat, Inc.
+# Copyright (C) 2021 Red Hat, Inc.
 #
 #    This program is free software: you can redistribute it and/or modify
 #    it under the terms of the GNU General Public License as published by
@@ -21,14 +21,13 @@
     Module used for building teflo scenario compound. Every teflo object
     has one scenario which has additional compounds associated to it.
 
-    :copyright: (c) 2020 Red Hat, Inc.
+    :copyright: (c) 2021 Red Hat, Inc.
     :license: GPLv3, see LICENSE for more details.
 """
 import errno
 import os
 import yaml
 from collections import OrderedDict
-from pykwalify.core import Core
 from pykwalify.errors import CoreError, SchemaError
 
 from .actions import Action
@@ -38,7 +37,7 @@ from .reports import Report
 from .notification import Notification
 from ..constants import SCENARIO_SCHEMA, SCHEMA_EXT, \
     SET_CREDENTIALS_OPTIONS
-from ..core import TefloResource
+from ..core import Inventory, TefloResource
 from ..exceptions import ScenarioError
 
 from ..helpers import gen_random_str, schema_validator
@@ -60,7 +59,10 @@ class Scenario(TefloResource):
         'name',         # the name of the scenario
         'description',  # a brief description of what the scenario is,
         'resource_check',    # external dependency check resources
-        'overall_status'
+        'overall_status',
+        'passed_tasks',
+        'failed_tasks',
+        'remote_workspace'
     ]
 
     def __init__(self,
@@ -68,6 +70,8 @@ class Scenario(TefloResource):
                  name=None,
                  parameters={},
                  validate_task_cls=ValidateTask,
+                 path: str = "",
+                 inventory_object: Inventory = Inventory,
                  **kwargs):
         """Constructor.
 
@@ -81,9 +85,13 @@ class Scenario(TefloResource):
         :type validate_task_cls: object
         :param kwargs: additional key:value(s)
         :type kwargs: dict
+        :param path: scenario file path
+        :type name: str
         """
         super(Scenario, self).__init__(config=config, name=name, **kwargs)
-
+        self._path = path
+        self._fullpath = ""
+        self._inventory: Inventory = inventory_object
         # set the scenario name attribute
         if not name:
             self._name = gen_random_str(15)
@@ -105,7 +113,15 @@ class Scenario(TefloResource):
         self._yaml_data = dict()
         # Properties to take care of included scenarios
         self._child_scenarios = list()
-        self._included_scenario_names = list()
+        self._included_scenario_path = list()
+        self._children_size = 0
+        self._my_parent = None
+
+        self._remote_workspace = list()
+        self._scenario_graph = None
+
+        self._passed_tasks = list()
+        self._failed_tasks = list()
 
         # set the teflo task classes for the scenario
         self._validate_task_cls = validate_task_cls
@@ -130,75 +146,156 @@ class Scenario(TefloResource):
         if parameters:
             self.load(parameters)
 
+    def __str__(self) -> str:
+        return "Scenario(\"%s\")" % self.name
+
     def get_all_resources(self):
         all_resources = list()
-        all_resources.extend([item for item in self.get_all_assets()])
-        all_resources.extend([item for item in self.get_all_actions()])
-        all_resources.extend([item for item in self.get_all_executes()])
-        all_resources.extend([item for item in self.get_all_reports()])
+
+        # The order is like below so we always execute resources in below order for a single scenario node
+
+        all_resources.extend([item for item in self.get_assets()])
+        all_resources.extend([item for item in self.get_actions()])
+        all_resources.extend([item for item in self.get_executes()])
+        all_resources.extend([item for item in self.get_reports()])
+        all_resources.extend([item for item in self.get_notifications()])
         return all_resources
 
-    def get_all_assets(self):
-        if self.child_scenarios:
-            all_assets = list()
-            for sc in self.child_scenarios:
-                all_assets.extend([item for item in getattr(sc, 'assets')])
-            all_assets.extend([item for item in getattr(self, 'assets')])
-            return all_assets
+    def get_assets(self):
         return getattr(self, 'assets')
 
-    def get_all_actions(self):
-        if self.child_scenarios:
-            all_actions = list()
-            for sc in self.child_scenarios:
-                all_actions.extend([item for item in getattr(sc, 'actions')])
-            all_actions.extend([item for item in getattr(self, 'actions')])
-            return all_actions
+    def get_actions(self):
         return getattr(self, 'actions')
 
-    def get_all_executes(self):
-        if self.child_scenarios:
-            all_executes = list()
-            for sc in self.child_scenarios:
-                all_executes.extend([item for item in getattr(sc, 'executes')])
-            all_executes.extend([item for item in getattr(self, 'executes')])
-            return all_executes
+    def get_executes(self):
         return getattr(self, 'executes')
 
-    def get_all_reports(self):
-        if self.child_scenarios:
-            all_reports = list()
-            for sc in self.child_scenarios:
-                all_reports.extend([item for item in getattr(sc, 'reports')])
-            all_reports.extend([item for item in getattr(self, 'reports')])
-            return all_reports
+    def get_reports(self):
         return getattr(self, 'reports')
 
-    def get_all_notifications(self):
-        if self.child_scenarios:
-            all_notifications = list()
-            for sc in self.child_scenarios:
-                all_notifications.extend([item for item in getattr(sc, 'notifications')])
-            all_notifications.extend([item for item in getattr(self, 'notifications')])
-            return all_notifications
+    def get_notifications(self):
         return getattr(self, 'notifications')
 
-    def add_resource(self, item):
+    def get_resource_idx(self, item):
+        """
+        Get the idx of the resource in its corresponding list.
+        :param item:
+        :return:
+        """
+        if isinstance(item, Asset):
+            if self._assets.__contains__(item):
+                return self._assets.index(item)
+            else:
+                return None
+        elif isinstance(item, Action):
+            if self._actions.__contains__(item):
+                return self._actions.index(item)
+            else:
+                return None
+        elif isinstance(item, Execute):
+            if self._executes.__contains__(item):
+                return self._executes.index(item)
+            else:
+                return None
+        elif isinstance(item, Report):
+            if self._reports.__contains__(item):
+                return self._reports.index(item)
+            else:
+                return None
+        elif isinstance(item, Notification):
+            if self._notifications.__contains__(item):
+                return self._notifications.index(item)
+            else:
+                return None
+        else:
+            raise ValueError('Resource must be of a valid Resource type.'
+                             'Check the type of the given item: %s' % item)
+
+    def replace_resource(self, item, idx):
+        """replace a scenario resource in its corresponding list.
+
+      :param item: resource
+      :type item: object
+      :param idx: index of resource list
+      :type idx: int
+      """
+        if isinstance(item, Asset):
+            self._assets[idx] = item
+        elif isinstance(item, Action):
+            self._actions[idx] = item
+        elif isinstance(item, Execute):
+            self._executes[idx] = item
+        elif isinstance(item, Report):
+            self._reports[idx] = item
+        elif isinstance(item, Notification):
+            self._notifications[idx] = item
+        else:
+            raise ValueError('Resource must be of a valid Resource type.'
+                             'Check the type of the given item: %s' % item)
+
+    def add_resource(self, item, idx=None):
         """Add a scenario resource to its corresponding list.
 
         :param item: resource data
         :type item: object
+        :param idx: index of resource list
+        :type idx: int
         """
         if isinstance(item, Asset):
-            self._assets.append(item)
+            if idx is not None:
+                cur_idx = self.get_resource_idx(item)
+                if cur_idx is not None:
+                    if idx != cur_idx:
+                        idx = cur_idx
+                    self.replace_resource(item=item, idx=idx)
+                else:
+                    self._assets.insert(idx, item)
+            else:
+                self._assets.append(item)
         elif isinstance(item, Action):
-            self._actions.append(item)
+            if idx is not None:
+                cur_idx = self.get_resource_idx(item)
+                if cur_idx is not None:
+                    if idx != cur_idx:
+                        idx = cur_idx
+                    self.replace_resource(item=item, idx=idx)
+                else:
+                    self._actions.insert(idx, item)
+            else:
+                self._actions.append(item)
         elif isinstance(item, Execute):
-            self._executes.append(item)
+            if idx is not None:
+                cur_idx = self.get_resource_idx(item)
+                if cur_idx is not None:
+                    if idx != cur_idx:
+                        idx = cur_idx
+                    self.replace_resource(item=item, idx=idx)
+                else:
+                    self._executes.insert(idx, item)
+            else:
+                self._executes.append(item)
         elif isinstance(item, Report):
-            self._reports.append(item)
+            if idx is not None:
+                cur_idx = self.get_resource_idx(item)
+                if cur_idx is not None:
+                    if idx != cur_idx:
+                        idx = cur_idx
+                    self.replace_resource(item=item, idx=idx)
+                else:
+                    self._reports.insert(idx, item)
+            else:
+                self._reports.append(item)
         elif isinstance(item, Notification):
-            self._notifications.append(item)
+            if idx is not None:
+                cur_idx = self.get_resource_idx(item)
+                if cur_idx is not None:
+                    if idx != cur_idx:
+                        idx = cur_idx
+                    self.replace_resource(item=item, idx=idx)
+                else:
+                    self._notifications.insert(idx, item)
+            else:
+                self._notifications.append(item)
         else:
             raise ValueError('Resource must be of a valid Resource type.'
                              'Check the type of the given item: %s' % item)
@@ -233,100 +330,76 @@ class Scenario(TefloResource):
 
     def reload_resources(self, tasks):
         """Reload scenario resources.
+
         This method is used to reload all resources after they have been processed (provisioned/orchestrate actions/
         execute tests/imported. Prior to adding the resource to the scenario, it filters the list by
         checking if the resource belongs to that scenario by comparing the names.
 
-        Then it filters the task list even more extracting the resource and the data in rvalue.
-        If rvalue is present then linchpin count feature was used and that more new host resources will be
-        created. If rvalue is not present then assume it is either a report resource or from a provisioning perspective
-        linchpin count was either not used (which is equivalent of count==1), or using old provisioners.
+        When filtering we extract the resource, the data in rvalue, and the index the reource is at
+        in the current resource list. If rvalue is present, then a provisioner with count
+        feature was used so more new host resources will be created dynamically as part of this reload.
+        If rvalue is not present then assume it is either a non-Asset resource or from a provisioning perspective
+        count was either not used (which is equivalent of count==1), or using old provisioners so the original
+        resource will be replaced with the updated resource in the task results lists.
 
         :param tasks: task data returned by blaster
         :type tasks: list
         """
-        count = 0
+
         filtered_task_list = list()
-        non_task_assets = list()
-        non_task_package = list()
 
-        # Filtering the resources based on labels. Separate steps for each resources, to make sure same name
-        # collisions dont happen between two different types of resources
+        # This a brute force way of resolving the out of order of resources when using labels.
+        # Prior to this change we would
+        # 1. filter task list for resource type the scenarios has ownership
+        # 2. filter existing resource list for the other resources of the same type that were filtered out by label
+        # 3. re-init the resource list for the resource type with an empty array
+        # 4. Append the resources that were updated from step 1
+        # 5. Then append the rest of the resources from step 2
+        # This lead to the resources being out of order in the results.yml
+        #
+        # Now we keep track of the index of the original resource so that we know where to replace/insert
+        # the new resources or updated resource in the list. This will support when the resource list is
+        # filtered using labels or status so we can retain proper ordering within the resource list during reload.
+        #
+        # New way
+        # 1. filter task list for scenario resource ownership and it's index location in resource list
+        # 2. if rvalue, then the first new resource will replace the original resource at the original index
+        # 3. if rvalue, for each 1+N rvalue will have the index incremented by 1 and inserted at the new index
+        #    pushing any resource that was at the index down the stack.
+        # 4. if not rvalue, the updated resource will check to see if the original resource is at the previous index
+        #    or at a new index (becaue it's been pushed down the stack) and replace the original resource accordingly.
 
+        # TODO: Simplify the filtering of resources when resources implement __eq__/__hash__
         filtered_task_list.extend(
-            [task for task in tasks if task.get('asset') and getattr(task.get('asset'), 'name')
+            [(task, self.get_resource_idx(task.get('asset')))
+             for task in tasks if task.get('asset') and getattr(task.get('asset'), 'name')
              in [h.name for h in self.assets]]
         )
         filtered_task_list.extend(
-            [task for task in tasks if isinstance(task.get('package'), Report) and (getattr(task.get('package'), 'name')
-             in [r.name for r in self.reports])]
+            [(task, self.get_resource_idx(task.get('package')))
+             for task in tasks if isinstance(task.get('package'), Report) and (getattr(task.get('package'), 'name')
+                                                                               in [r.name for r in self.reports])]
         )
 
         filtered_task_list.extend(
-            [task for task in tasks if
+            [(task, self.get_resource_idx(task.get('package'))) for task in tasks if
              isinstance(task.get('package'), Execute) and (getattr(task.get('package'), 'name')
                                                            in [r.name for r in self.executes])]
         )
 
         filtered_task_list.extend(
-            [task for task in tasks if isinstance(task.get('package'), Action) and (getattr(task.get('package'), 'name')
-             in [r.name for r in self.actions])]
+            [(task, self.get_resource_idx(task.get('package')))
+             for task in tasks if isinstance(task.get('package'), Action) and (getattr(task.get('package'), 'name')
+                                                                               in [r.name for r in self.actions])]
         )
 
-        # using labels in SDF will have only specific resources selected. So collecting the non task related
-        # assets and report resources to be added back to the scenario resources. This is being done in order to
-        # put the non provisioned asset and non imported report resources into results.yml
-        non_task_assets.extend([asset for asset in self.assets if asset.name not in
-                                [getattr(task.get('asset'), 'name') for task in filtered_task_list
-                                 if task.get('asset')]])
-
-        non_task_package.extend([report for report in self.reports if report.name not in
-                                 [getattr(task.get('package'), 'name') for task in filtered_task_list if
-                                  isinstance(task.get('package'), Report) and (getattr(task.get('package'), 'name'))]])
-
-        non_task_package.extend([execute for execute in self.executes if execute.name not in
-                                 [getattr(task.get('package'), 'name') for task in filtered_task_list if
-                                  isinstance(task.get('package'), Execute) and (getattr(task.get('package'), 'name'))]])
-
-        non_task_package.extend([action for action in self.actions if action.name not in
-                                 [getattr(task.get('package'), 'name') for task in filtered_task_list if
-                                  isinstance(task.get('package'), Action) and (getattr(task.get('package'), 'name'))]])
-
-        for res, rvalue in [(res, item['rvalue']) for task in filtered_task_list
-                            for res in [task.get(r) for r in ['asset', 'package'] if r in task.keys()]
-                            for item in task.get('methods')]:
-            if count == 0:
-                if rvalue is not None:
-                    # initialize the host resource list to remove any previous host resources
-                    self.initialize_resource(res)
-                    # load the new host resources using the parameters from item['rvalue']
-                    self.load_resources(Asset, rvalue)
-                    count += 1
-                else:
-                    self.initialize_resource(res)
-                    self.add_resource(res)
-                    count += 1
+        for res, rvalue, idx in [(res, item['rvalue'], idx) for task, idx in filtered_task_list
+                                 for res in [task.get(r) for r in ['asset', 'package'] if r in task.keys()]
+                                 for item in task.get('methods')]:
+            if rvalue is not None:
+                self.load_resources(Asset, rvalue, idx)
             else:
-                if rvalue is not None:
-                    self.load_resources(Asset, rvalue)
-                else:
-                    self.add_resource(res)
-
-        if count > 0:
-            task_keys = list()
-            task_keys.extend([k for task in tasks for k in task.keys()])
-            if 'asset' in task_keys:
-                # Adding assets which were not part of the labels used
-                for res in non_task_assets:
-                    self.add_resource(res)
-            elif 'package' in task_keys:
-                # Adding other resources except asset which were not part of the labels used
-                # getting the type of resource is the tasks list and comparing it with the non_task_resources
-                # if they are of same type then add the resource else pass
-                pkg_type = type(tasks[0].get('package'))
-                for res in non_task_package:
-                    if isinstance(res, pkg_type):
-                        self.add_resource(res)
+                self.add_resource(res, idx)
 
         return
 
@@ -431,6 +504,34 @@ class Scenario(TefloResource):
             raise ValueError('Execute must be of type %s ' % type(Execute))
         self._executes.append(execute)
 
+    @property
+    def included_scenario_path(self):
+        """scenario path property.
+
+        :return: scenario path property to the scenario
+        :rtype: list
+        """
+        return self._included_scenario_path
+
+    @included_scenario_path.setter
+    def included_scenario_path(self, included_scenario_path):
+        """scenario path property."""
+        self._included_scenario_path.append(included_scenario_path)
+
+    @property
+    def remote_workspace(self):
+        """remote_workspace property.
+
+        :return: remote_workspace property to the scenario
+        :rtype: list
+        """
+        return self._remote_workspace
+
+    @remote_workspace.setter
+    def remote_workspace(self, remote_workspace):
+        """Set remote_workspace property."""
+        self._remote_workspace = remote_workspace
+
     @property
     def reports(self):
         """Reports property.
@@ -446,6 +547,46 @@ class Scenario(TefloResource):
         raise ValueError('You can not set reports directly.'
                          'Use function ~Scenario.add_reports')
 
+    @property
+    def path(self):
+        return self._path
+
+    @path.setter
+    def path(self, path):
+        self._path = path
+
+    @property
+    def fullpath(self):
+        return self._fullpath
+
+    @fullpath.setter
+    def fullpath(self, fullpath):
+        self._fullpath = fullpath
+
+    @property
+    def inventory(self):
+        return self._inventory
+
+    @inventory.setter
+    def inventory(self, inventory: Inventory):
+        self._inventory = inventory
+
+    @property
+    def children_size(self):
+        return self._children_size
+
+    @children_size.setter
+    def children_size(self, children_size):
+        self._children_size = children_size
+
+    @property
+    def my_parent(self):
+        return self._my_parent
+
+    @my_parent.setter
+    def my_parent(self, my_parent):
+        self._my_parent = my_parent
+
     @property
     def child_scenarios(self):
         """
@@ -469,27 +610,6 @@ class Scenario(TefloResource):
             raise ValueError('scenario must be of type %s ' % type(Scenario))
         self._child_scenarios.append(scenario)
 
-    @property
-    def included_scenario_names(self):
-        """
-        List of files which are put in the include section of the scenario
-        :return: list of files
-        :rtype: list
-        """
-        return self._included_scenario_names
-
-    @included_scenario_names.setter
-    def included_scenario_names(self, inc_name_list):
-        """Add included file names as list
-        :param inc_name_list: list of names
-        :type inc_name_list: list
-        """
-        if inc_name_list:
-            for item in inc_name_list:
-                self.included_scenario_names.append(item)
-        else:
-            raise ValueError("Included scenario list cannot be empty")
-
     def add_reports(self, report):
         """Add report resources to the scenario.
 
@@ -515,6 +635,44 @@ class Scenario(TefloResource):
         raise ValueError('You can not set notifications directly.'
                          'Use function ~Scenario.add_notifications')
 
+    @property
+    def scenario_graph(self):
+        """scenario_graph property for the resource"""
+        return self._scenario_graph
+
+    @scenario_graph.setter
+    def scenario_graph(self, value):
+        """set scenario_graph property"""
+        self._scenario_graph = value
+
+    @property
+    def passed_tasks(self):
+        """passed_tasks property for the scenario"""
+        return self._passed_tasks
+
+    @passed_tasks.setter
+    def passed_tasks(self, value):
+        """set passed_tasks property"""
+        if not isinstance(value, list):
+            raise ScenarioError("The scenario passed_tasks value needs to be a list")
+        for task in value:
+            if task not in self.passed_tasks:
+                self._passed_tasks.append(task)
+
+    @property
+    def failed_tasks(self):
+        """failed_tasks property for the scenario"""
+        return self._failed_tasks
+
+    @failed_tasks.setter
+    def failed_tasks(self, value):
+        """set failed_tasks property"""
+        if not isinstance(value, list):
+            raise ScenarioError("The scenario failed_tasks value needs to be a list")
+        for task in value:
+            if task not in self.failed_tasks:
+                self._failed_tasks.append(task)
+
     def add_notifications(self, notification):
         """Add notifications resources to the scenario.
 
@@ -554,17 +712,18 @@ class Scenario(TefloResource):
         profile = OrderedDict()
         profile['name'] = self.name
         profile['description'] = self.description
+        profile['remote_workspace'] = self.remote_workspace
         if self.child_scenarios:
-            profile['include'] = self.included_scenario_names
+            profile['include'] = self.included_scenario_path
         profile['resource_check'] = self.resource_check
         profile['provision'] = [asset.profile() for asset in self.assets]
         profile['orchestrate'] = [action.profile() for action in self.actions]
         profile['execute'] = [execute.profile() for execute in self.executes]
         profile['report'] = [report.profile() for report in self.reports]
         profile['notifications'] = [notification.profile() for notification in self.notifications]
-        if hasattr(self, 'overall_status'):
-            profile['overall_status'] = getattr(self, 'overall_status')
-
+        for prop in ['overall_status', 'passed_tasks', 'failed_tasks']:
+            if hasattr(self, prop):
+                profile[prop] = getattr(self, prop)
         return profile
 
     def _construct_validate_task(self):
@@ -581,7 +740,7 @@ class Scenario(TefloResource):
         }
         return task
 
-    def load_resources(self, res_type, res_list):
+    def load_resources(self, res_type, res_list, idx=None):
         """
         Load the resource in the scenario list of `res_type`.
 
@@ -598,13 +757,25 @@ class Scenario(TefloResource):
         :param res_type: The type of resources the function will load into its
             list
         :param res_list: A list of resources dict
+        :param idx: integer index where resource will be replaced/insert
         :return: None
         """
         # No resources defined, then exit
         if not res_list:
             return
-
+        increment = False
         for item in res_list:
-            self.add_resource(
-                res_type(config=self.config,
-                         parameters=item))
+            if idx is not None and not increment:
+                self.replace_resource(item=res_type(config=self.config, parameters=item), idx=idx)
+                increment = True
+                continue
+            if idx is not None and increment:
+                idx += 1
+            self.add_resource(item=res_type(config=self.config, parameters=item), idx=idx)
+
+    def graph_str(self, level=0):
+        ret = "   " * level + "->" + self.path + "\n"
+        child: Scenario
+        for child in self.child_scenarios:
+            ret += child.graph_str(level + 1)
+        return ret
