@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2020 Red Hat, Inc.
+# Copyright (C) 2021 Red Hat, Inc.
 #
 #    This program is free software: you can redistribute it and/or modify
 #    it under the terms of the GNU General Public License as published by
@@ -20,22 +20,32 @@
 
     Teflo a framework to test product interoperability.
 
-    :copyright: (c) 2020 Red Hat, Inc.
+    :copyright: (c) 2021 Red Hat, Inc.
     :license: GPLv3, see LICENSE for more details.
 """
+from copy import deepcopy
 import errno
+from logging import root
 import os
+from re import L
 import sys
-
+from teflo.exceptions import TefloScenarioFailure
 import blaster
+import termcolor
+from contextlib import ExitStack
+from functools import partial
 import yaml
-from glob import glob
+import click
+import shutil
+from distutils.dir_util import copy_tree
+from teflo.helpers import exec_local_cmd
 from . import __name__ as __teflo_name__
-from .constants import TASKLIST, RESULTS_FILE, DATA_FOLDER, DEFAULT_INVENTORY, DEFAULT_ARTIFACT
+from .constants import NOTIFYSTATES, TASKLIST, RESULTS_FILE, DATA_FOLDER, DEFAULT_INVENTORY
 from .core import TefloError, LoggerMixin, TimeMixin, Inventory
-from .helpers import file_mgmt, gen_random_str, sort_tasklist
+from .helpers import file_mgmt, gen_random_str, sort_tasklist, preproc_path
 from .resources import Scenario, Asset, Action, Report, Execute, Notification
 from .utils.config import Config
+from .utils.scenario_graph import ScenarioGraph
 from .utils.pipeline import PipelineFactory
 
 
@@ -105,6 +115,11 @@ class Teflo(LoggerMixin, TimeMixin):
                 self._teflo_options['skip_notify'] = value
             if key == 'no_notify' and value:
                 self._teflo_options['no_notify'] = value
+            # assigning cli iterate_method value to self.config['INCLUDED_SDF_ITERATE_METHOD']
+            if key == 'iterate_method' and value:
+                self.config['INCLUDED_SDF_ITERATE_METHOD'] = value
+            if key == 'skip_fail':
+                self._teflo_options['skip_fail'] = value
 
         if log_level:
             self.config['LOG_LEVEL'] = log_level
@@ -191,9 +206,15 @@ class Teflo(LoggerMixin, TimeMixin):
         self.config['WORKSPACE'] = self.workspace
 
         # creating one time inventory object
-        self.cbn_inventory = Inventory.get_instance(self.config, self._uid)
+        self.cbn_inventory: Inventory = Inventory.get_instance(self.config, self._uid)
 
-        self.scenario = Scenario(config=self.config)
+        self.scenario_graph = ScenarioGraph()
+
+        # Updating the os.environ variables with TEFLO_DATA_FOLDER and TEFLO_RESULTS_FOLDER
+        os.environ.update({'TEFLO_DATA_FOLDER': os.path.abspath(self.config.get('DATA_FOLDER')),
+                           'TEFLO_RESULTS_FOLDER': os.path.abspath(self.config.get('RESULTS_FOLDER')),
+                           'TEFLO_WORKSPACE': os.path.abspath(self.config.get('WORKSPACE'))
+                           })
 
     @property
     def name(self):
@@ -226,14 +247,13 @@ class Teflo(LoggerMixin, TimeMixin):
     def teflo_options(self):
         return self._teflo_options
 
-    def _populate_scenario_resources(self, scenario_obj, scenario_stream):
+    def _populate_scenario_resources(self, scenario_obj: Scenario, scenario_stream):
 
         scenario_data = yaml.safe_load(scenario_stream)
         pro_items = scenario_data.pop('provision', None)
         orc_items = scenario_data.pop('orchestrate', None)
         exe_items = scenario_data.pop('execute', None)
         rpt_items = scenario_data.pop('report', None)
-        inc_items = scenario_data.pop('include', None)
         notify_items = scenario_data.pop('notifications', None)
 
         scenario_obj.load(scenario_data)
@@ -244,43 +264,23 @@ class Teflo(LoggerMixin, TimeMixin):
         scenario_obj.load_resources(Report, rpt_items)
         scenario_obj.load_resources(Notification, notify_items)
 
-    def load_from_yaml(self, filedata):
+    def load_from_yaml(self, scenario_graph: ScenarioGraph):
         """
-        Given the YAML filename with the scenario description, this
-        function will load all resources described in the file with
-        the teflo object.
-
-        The main sections of the YAML descriptor (provision, orchestrate,
-        execute and report) will be taken of the data in its respective
-        data object. All the rest of the attributes within the YAML file
-        will be loaded into the ~self.scenario object.
-
-        If include section for included scenarios is present then they get
-        added as child scenario objects for the main scenario(~self.scenario).
-        These scenarios in turn go through the same process of taking out the
-        main sections from the data in its respective data object
-
-        The additional attributes within the file will be ignored by the Scenario
-        resource.
-
-        Once each data section is taken of the main descriptor, it will
-        then be be loaded in each respective lists within the
-        ~self.scenario object.
-
-        :param filedata: list of full data object for the YAML file descriptor
-        :return:
+        This method load and populate all resources from
+        the scenario graph, it also validate all labels
+        for all nodes inside of the scenaio graph
         """
         # setting up master scenario
-        self.scenario.yaml_data = filedata[0]
-        self._populate_scenario_resources(self.scenario, filedata[0])
+        self.scenario_graph = scenario_graph
 
-        # creating child scenario objects
-        if len(filedata) > 1:
-            for scenario_stream in filedata[1:]:
-                ch_scenario = Scenario(config=self.config)
-                ch_scenario.yaml_data = scenario_stream
-                self._populate_scenario_resources(ch_scenario, scenario_stream)
-                self.scenario.add_child_scenario(ch_scenario)
+        # Iterate using certain iterator, this is post-order traversal for a tree-liked graph
+        sc: Scenario
+        for sc in self.scenario_graph:
+            # Add resources to sc objects
+            self._populate_scenario_resources(sc, sc.yaml_data)
+            self.scenario_graph.reload_resources_from_scenario(sc)
+            # setting scenario_graph property to each scenario in the scenario_graph
+            sc.__setattr__('scenario_graph', self.scenario_graph)
 
         self._validate_labels()
 
@@ -288,9 +288,9 @@ class Teflo(LoggerMixin, TimeMixin):
         """This method displays all the labels in the scenario"""
         res_label_dict = dict()
         # getting all labels from all resources in the scenario
-        for res in self.scenario.get_all_resources():
+        for res in self.scenario_graph.get_all_resources():
             res_label_dict[res] = dict(name=getattr(res, 'name'),
-                                       labels=[lab for lab in getattr(res, 'labels')])
+                                        labels=[lab for lab in getattr(res, 'labels')])
         self.logger.info('-' * 79)
         self.logger.info('SCENARIO LABELS'.center(79))
         self.logger.info('-' * 79)
@@ -334,7 +334,8 @@ class Teflo(LoggerMixin, TimeMixin):
             res_label_list = list()
             user_labels = list()
             # labels present in all res
-            res_label_list.extend([lab for res in self.scenario.get_all_resources() for lab in getattr(res, 'labels')])
+            res_label_list.extend(
+                [lab for sc in self.scenario_graph for res in sc.get_all_resources() for lab in getattr(res, 'labels')])
             # labels provided by user at cli
             user_labels.extend([lab for lab in
                                 self.teflo_options.get('labels', []) or self.teflo_options.get('skip_labels', [])])
@@ -347,7 +348,56 @@ class Teflo(LoggerMixin, TimeMixin):
                                       "scenario descriptor file" % label)
         return
 
-    def run(self, tasklist=TASKLIST):
+    def run_all_validate_helper(self, final_passed_tasks, final_failed_tasks, status):
+        """
+        This is a helper method for running validate task for the whole scenario graph
+        """
+        do_validate = True
+        self.logger.info(termcolor.colored("Validating against all scenarios!", "green"))
+        self.logger.info("." * 50)
+        sc: Scenario
+        for sc in self.scenario_graph:
+            if do_validate:
+                self.logger.info("." * 50)
+                self.logger.info(termcolor.colored('\'%s\' is validated from the scenario file: %s' %
+                                                    (sc.name, sc.path), "green"))
+                self.logger.info("." * 50)
+                self.run_helper(sc=sc, tasklist=["validate"])
+            if getattr(sc, "overall_status", None) is not None:
+                if getattr(sc, "overall_status") == 1:
+                    do_validate = False
+        self.collect_final_passed_failed_tasks_status(final_passed_tasks, final_failed_tasks, status)
+
+    def run_all_helper(self, tasklist: list, final_passed_tasks: list, final_failed_tasks: list, status: list):
+        """
+        This is a helper method for running all tasks for the whole scenario graph
+        it will update final_passed_tasks, final_failed_tasks and status after it
+        finish the run
+        """
+
+        with ExitStack() as stack:
+
+            tasklist_run: list = deepcopy(tasklist)
+            if "validate" in tasklist:
+                tasklist_run.remove("validate")
+                self.run_all_validate_helper(final_passed_tasks, final_failed_tasks, status)
+                if status[0] == 1:
+                    return
+            if "cleanup" in tasklist:
+                tasklist_run.remove("cleanup")
+                stack.callback(partial(self.cleanup_helper, final_passed_tasks, final_failed_tasks, status))
+
+            sc: Scenario
+            for sc in self.scenario_graph:
+                try:
+                    self.run_helper(sc=sc, tasklist=tasklist_run)
+                except TefloScenarioFailure as ex:
+                    self.scenario_graph.reinit()
+                    self.logger.error(ex)
+                    break
+            self.collect_final_passed_failed_tasks_status(final_passed_tasks, final_failed_tasks, status)
+
+    def run(self, tasklist: list = TASKLIST):
         """
         This function assumes there are zero or more tasks to be
         loaded within the list of pipelines: ~self.pipelines.
@@ -361,46 +411,143 @@ class Teflo(LoggerMixin, TimeMixin):
         pipeline and then each pipeline is sent to blaster blastoff.
         For every pipeline within ~self.pipelines,
         """
-        # lists to control which tasks passed or failed
-        passed_tasks = list()
-        failed_tasks = list()
 
+        # initialize the final passed task and failed task
+        final_passed_tasks = []
+        final_failed_tasks = []
         # initialize overall status
-        status = 0
+        status = [0]
 
         # save start time
         self.start()
 
         self._print_header(tasklist)
 
+        self.run_all_helper(tasklist, final_passed_tasks, final_failed_tasks, status)
+
+        for task in final_failed_tasks:
+            if task in final_passed_tasks:
+                final_passed_tasks.remove(task)
+
+        self.end()
+        # determine state
+        # 0 is false here
+        state = 'FAILED' if status[0] else 'PASSED'
+
+        self._write_out_results()
+
+        self._print_footer(final_passed_tasks, final_failed_tasks, state)
+
+        self._archive_results()
+
+        if self.config["CLEAN_CACHED_WORKSPACE_AFTER_EACH_RUN"].lower() == "true":
+            self.clean_cached_remote_workspace()
+
+        sys.exit(status[0])
+
+    def clean_cached_remote_workspace(self):
+        if os.path.isdir(self.config["REMOTE_WORKSPACE_DOWNLOAD_LOCATION"]):
+            shutil.rmtree(self.config["REMOTE_WORKSPACE_DOWNLOAD_LOCATION"])
+
+    def cleanup_helper(self, final_passed_tasks: list, final_failed_tasks: list, status: list):
+        """
+        This is a helper method for cleanup. It does cleanup for all scenarios in
+        the scenario graph. It does cleanup in a reverse order to the provisoin/
+        orchestarte/execute/report
+        """
+
+        cleanup_sc = []
+        for sc in self.scenario_graph:
+            cleanup_sc.append(sc)
+        cleanup_sc.reverse()
+        sc: Scenario
+        for sc in cleanup_sc:
+            try:
+                self.logger.info("." * 50)
+                self.logger.info(termcolor.colored('\'%s\' is running from the scenario file: %s' %
+                                                   (sc.name, sc.path), "green"))
+                self.logger.info("." * 50)
+                self.run_helper(sc=sc, tasklist=["cleanup"])
+            except TefloScenarioFailure as ex:
+                self.scenario_graph.reinit()
+                self.logger.error(ex)
+                break
+
+        self.collect_final_passed_failed_tasks_status(final_passed_tasks, final_failed_tasks, status)
+
+    def collect_final_passed_failed_tasks_status(self, final_passed_tasks: list,
+                                                 final_failed_tasks: list, status: list):
+        """
+        This method collects all failed and passed tasks from all scenarios in
+        the self.scenario_graph
+        """
+
+        for sc in self.scenario_graph:
+
+            if getattr(sc, "failed_tasks", None) is not None:
+                for task in sc.failed_tasks:
+                    if task not in final_failed_tasks:
+                        final_failed_tasks.append(task)
+                    if task in final_passed_tasks:
+                        final_passed_tasks.remove(task)
+
+            if getattr(sc, "passed_tasks", None) is not None:
+                for task in sc.passed_tasks:
+                    if task not in final_passed_tasks and task not in final_failed_tasks:
+                        final_passed_tasks.append(task)
+
+            if getattr(sc, "overall_status", None) is not None:
+                if sc.overall_status == 1:
+                    status[0] = 1
+
+    def run_helper(self, sc: Scenario = None, tasklist: list = TASKLIST):
+        """
+        This is a helper method for running all tasks for a certain scenario
+        """
+        passed_tasks = list()
+        failed_tasks = list()
+        if len(tasklist) != 0:
+            self.logger.info("." * 50)
+            self.logger.info(termcolor.colored('\'%s\' is running from the scenario file: %s' %
+                                               (sc.name, sc.path), "green"))
+            self.logger.info("." * 50)
+        # initialize overall status
+        status = 0
+        if not self.teflo_options.get('no_notify', False):
+            self.notify('on_start', status, passed_tasks, failed_tasks, scenario=sc)
+
         try:
             for task in sort_tasklist(tasklist):
                 self.logger.info(' * Task    : %s' % task)
 
                 # initially update list of passed tasks
                 passed_tasks.append(task)
-                if not self.teflo_options.get('no_notify', False):
-                    self.notify('on_start', status, passed_tasks, failed_tasks)
-                data = self._run_pipeline(task)
+                # TODO: MAKE THIS ONE TIME RUN
+
+                data = self._run_pipeline(task, sc)
+                self.scenario_graph.remove_resources_from_scenario(sc)
 
                 # reload resource objects
-                self.scenario.reload_resources(data)
-
-                if self.scenario.child_scenarios:
-                    [sc.reload_resources(data) for sc in self.scenario.child_scenarios]
-
+                sc: Scenario
+                # for sc in self.scenario_graph:
+                sc.reload_resources(data)
+                self.scenario_graph.reload_resources_from_scenario(sc)
                 # Creating inventory only when task is provision
                 if task == 'provision':
-                    try:
-                        # create the master inventory
-                        for host in self.scenario.get_all_assets():
-                            if (hasattr(host, 'role') or hasattr(host, 'groups')) and hasattr(host, 'ip_address'):
-                                self.logger.info('Populating master inventory file with host(s) %s'
-                                                 % getattr(host, 'name'))
+                    all_hosts = sc.get_assets()
 
-                        self.cbn_inventory.create_master(all_hosts=self.scenario.get_all_assets())
-                    except Exception as ex:
-                        raise TefloError("Error while creating the master inventory %s" % ex)
+                    if all_hosts:
+                        for host in all_hosts:
+                            if hasattr(host, 'ip_address'):
+                                self.logger.info('Populating inventory file with host(s) %s'
+                                                    % getattr(host, 'name'))
+                        try:
+                            self.cbn_inventory.create_inventory(all_hosts=all_hosts)
+
+                        except Exception as ex:
+                            raise TefloError("Error while creating the inventory for scenario %s: %s" % (sc.path, ex))
+                    else:
+                        self.logger.info("No hosts provisioned to be added to the inventory")
 
                 self.logger.info("." * 50)
         except Exception as ex:
@@ -415,111 +562,129 @@ class Teflo(LoggerMixin, TimeMixin):
 
             self.logger.error(ex)
 
+            self.scenario_graph.remove_resources_from_scenario(sc)
             # reload resource objects
-            self.scenario.reload_resources(ex.results)
-
-            if self.scenario.child_scenarios:
-                [sc.reload_resources(ex.results) for sc in self.scenario.child_scenarios]
-
+            sc.reload_resources(ex.results)
+            self.scenario_graph.reload_resources_from_scenario(sc)
             # roll back by cleaning up any resources that might have been provisioned
-            if 'cleanup' in tasklist and [item for item in failed_tasks if item != 'cleanup']:
+            if "cleanup" in tasklist and [item for item in failed_tasks if item != 'cleanup']:
                 if [item for item in passed_tasks if item == 'provision'] \
                         or [item for item in failed_tasks if item == 'provision']:
                     self.logger.info("\n\n")
                     self.logger.warning("A failure occurred before running the cleanup task. "
                                         "Attempting to run the cleanup task to roll back all provisioned resources.")
-                    task = tasklist[tasklist.index('cleanup')]
 
                     try:
-
-                        data = self._run_pipeline(task)
+                        data = self._run_pipeline("cleanup", sc)
+                        self.scenario_graph.remove_resources_from_scenario(sc)
 
                         # reload resource objects
-                        self.scenario.reload_resources(data)
-                        if self.scenario.child_scenarios:
-                            [sc.reload_resources(data) for sc in self.scenario.child_scenarios]
+                        sc.reload_resources(data)
                         passed_tasks.append(task)
                     except Exception as ex:
                         self.logger.error(ex)
                         self.logger.error("There was a problem running the cleanup task to roll back the "
                                           "resources. You may need to manually cleanup any provisioned resources")
                         failed_tasks.append(task)
-                        raise
+                        raise TefloScenarioFailure("`%s` failed during running cleanup rollback" % sc.name)
         finally:
-            # save end time
-            self.end()
-            # determine state
-            state = 'FAILED' if status else 'PASSED'
-
-            # finally send out any notifications
+            setattr(sc, 'overall_status', status)
+            setattr(sc, 'passed_tasks', passed_tasks)
+            setattr(sc, 'failed_tasks', failed_tasks)
+            self.scenario_graph.reload_resources_from_scenario(sc)
             if not self.teflo_options.get('no_notify', False):
-                self.notify('on_complete', status, passed_tasks, failed_tasks)
+                self.notify('on_complete', status, passed_tasks, failed_tasks, sc)
 
-            self._write_out_results()
+            def exit_on_status():
 
-            self._print_footer(passed_tasks, failed_tasks, state)
+                state = 'FAILED' if status else 'PASSED'
 
-            self._archive_results()
+                if str(self.config.get("SKIP_FAIL")).lower() != 'true' and \
+                        str(self.config.get("SKIP_FAIL")).lower() != 'false' and \
+                        self.config.get("SKIP_FAIL") is not None:
+                    self.logger.error('The skip_fail variable can be true or false only. Running'
+                                      ' as default value which is False')
+                    self.config["SKIP_FAIL"] = 'False'
 
-            sys.exit(status)
+                if str(self.config.get("SKIP_FAIL")).lower() != 'true' \
+                        and self._teflo_options.get('skip_fail') is not True and state is 'FAILED':
+                    raise TefloScenarioFailure(
+                        "Scenario `%s` failed during the Teflo run" % sc.name)
 
-    def notify(self, task, status=0, passed_tasks=None, failed_tasks=None):
+            exit_on_status()
 
-        setattr(self.scenario, 'overall_status', status)
-        setattr(self.scenario, 'passed_tasks', [])
-        setattr(self.scenario, 'failed_tasks', [])
+    def notify(self, task, status=0, passed_tasks=None, failed_tasks=None, scenario: Scenario = None):
+        """
+        This method handle all notifications
+        """
 
-        if passed_tasks:
-            setattr(self.scenario, 'passed_tasks', passed_tasks)
-
-        if failed_tasks:
-            setattr(self.scenario, 'failed_tasks', failed_tasks)
+        self.logger.debug('Sending out any notifications that are registered.')
 
         if task == 'on_demand':
             self.start()
             self._print_header(['notify'])
             self.logger.info(' * Task    : notify')
+            sc: Scenario
+            for sc in self.scenario_graph:
+                try:
+                    data = self._run_pipeline(task, sc)
+                    self.scenario_graph.remove_resources_from_scenario(sc)
+                    sc.reload_resources(data)
+                    self.scenario_graph.reload_resources_from_scenario(sc)
+                except Exception as ex:
+                    status = 1
+                    self.logger.error(ex)
+                    self.logger.error(termcolor.colored(
+                        'One or more notifications failed. Refer to the scenario.log', "red"))
+                    self.scenario_graph.remove_resources_from_scenario(sc)
+                    sc.reload_resources(ex.results)
+                    self.scenario_graph.reload_resources_from_scenario(sc)
+                    # save end time
+            self.end()
+            # determine state
+            state = 'FAILED' if status else 'PASSED'
 
-        # blast off the pipeline list of tasks
-        try:
+            self._write_out_results()
+            final_passed_tasks = []
+            final_failed_tasks = []
+            self.collect_final_passed_failed_tasks_status(final_passed_tasks, final_failed_tasks, [status])
 
-            self.logger.info('Sending out any notifications that are registered.')
-            data = self._run_pipeline(task)
+            self._print_footer(final_passed_tasks, final_failed_tasks, state)
 
-            # reload resource objects
-            self.scenario.reload_resources(data)
+            self._archive_results()
+            sys.exit(status)
+        else:
+            setattr(scenario, 'overall_status', status)
 
-            if self.scenario.child_scenarios:
-                [sc.reload_resources(data) for sc in self.scenario.child_scenarios]
-        except Exception as ex:
-            status = 1
-            self.logger.error(ex)
-            self.logger.error('One or more notifications failed. Refer to the scenario.log')
+            if passed_tasks:
+                setattr(scenario, 'passed_tasks', passed_tasks)
 
-            # reload resource objects
-            self.scenario.reload_resources(ex.results)
+            if failed_tasks:
+                setattr(scenario, 'failed_tasks', failed_tasks)
 
-            if self.scenario.child_scenarios:
-                [sc.reload_resources(ex.results) for sc in self.scenario.child_scenarios]
-        finally:
-            if task == 'on_demand':
-                # save end time
-                self.end()
-                # determine state
-                state = 'FAILED' if status else 'PASSED'
+            # blast off the pipeline list of tasks
+            try:
+                data = self._run_pipeline(task, scenario)
+                self.scenario_graph.remove_resources_from_scenario(scenario)
+                scenario.reload_resources(data)
+                self.scenario_graph.reload_resources_from_scenario(scenario)
+            except Exception as ex:
+                status = 1
+                self.logger.error(ex)
+                self.logger.error('One or more notifications failed. Refer to the scenario.log')
+                self.scenario_graph.remove_resources_from_scenario(scenario)
+                scenario.reload_resources(ex.results)
+                self.scenario_graph.reload_resources_from_scenario(scenario)
 
-                self._write_out_results()
+    @property
+    def root_scenario_filename(self):
+        return self.scenario_graph.root.path
 
-                self._print_footer(getattr(self.scenario, 'passed_tasks'),
-                                   getattr(self.scenario, 'failed_tasks'),
-                                   state)
-
-                self._archive_results()
-
-                sys.exit(status)
-
-    def _run_pipeline(self, task):
+    @property
+    def root_scenario_fullpath(self):
+        return self.scenario_graph.root.fullpath
 
+    def _run_pipeline(self, task, scenario: Scenario):
         data = {}
 
         # create a pipeline builder object
@@ -530,20 +695,24 @@ class Teflo(LoggerMixin, TimeMixin):
             self.logger.warning('Task %s is not valid by teflo.', task)
             return data
 
-        pipeline = pipe_builder.build(self.scenario, self.teflo_options)
-
-        self.logger.info('.' * 50)
-        self.logger.info('Starting tasks on pipeline: %s',
-                         pipeline.name)
-
-        # check if pipeline has tasks to be run
-        if not pipeline.tasks:
-            self.logger.warning('... no tasks to be executed ...')
+        pipeline = pipe_builder.build(scenario, self.teflo_options, scenario_graph=self.scenario_graph)
+        if pipeline.name == 'notify' and not scenario.notifications:
+            self.logger.debug('.' * 50)
+            self.logger.debug('Starting tasks on pipeline: %s',
+                              pipeline.name)
+            self.logger.debug('... no tasks to be executed ...')
             return data
+        else:
+            self.logger.info('.' * 50)
+            self.logger.info('Starting tasks on pipeline: %s',
+                             pipeline.name)
+            # check if pipeline has tasks to be run
+            if not pipeline.tasks:
+                self.logger.warning('... no tasks to be executed ...')
+                return data
 
         # create blaster object with pipeline to run
         blast = blaster.Blaster(pipeline.tasks)
-
         # blast off the pipeline list of tasks reload_resources
         data = blast.blastoff(
             serial=not pipeline.type.__concurrent__,
@@ -552,6 +721,40 @@ class Teflo(LoggerMixin, TimeMixin):
 
         return data
 
+    @property
+    def data_folder_results_yml_path(self):
+        return os.path.join(self.data_folder, RESULTS_FILE)
+
+    @property
+    def final_results_yml_path(self):
+        return os.path.join(self.config["RESULTS_FOLDER"], RESULTS_FILE)
+
+    def _data_folder_resultsyml_exist(self):
+        return os.path.exists(self.data_folder_results_yml_path)
+
+    def _final_results_yml_exists(self):
+        return os.path.exists(self.final_results_yml_path)
+
+    def _create_final_result_yml(self):
+        """
+        This method create the final result
+        scenario file and put it to data folder
+        """
+
+        if self._data_folder_resultsyml_exist():
+            return
+        if preproc_path(self.config["RESULTS_FOLDER"]) in self.root_scenario_fullpath or \
+            preproc_path(self.data_folder) in self.root_scenario_fullpath or \
+                self.root_scenario_filename == "results.yml":
+            root_scenario_results_path = os.path.join(self.data_folder, self.root_scenario_filename)
+        else:
+            root_scenario_results_path = os.path.join(
+                self.data_folder, self.root_scenario_filename.split(".")[0] + "_results.yml")
+        if root_scenario_results_path != self.final_results_yml_path and self.root_scenario_filename != "results.yml":
+            if self._final_results_yml_exists():
+                os.remove(self.final_results_yml_path)
+        os.system("cp %s %s" % (root_scenario_results_path, self.data_folder_results_yml_path))
+
     def _print_header(self, tasklist):
 
         self.logger.info('\n')
@@ -561,29 +764,32 @@ class Teflo(LoggerMixin, TimeMixin):
         self.logger.info(' * Workspace             : %s' % self.workspace)
         self.logger.info(' * Log Level             : %s' % self.config['LOG_LEVEL'])
         self.logger.info(' * Tasks                 : %s' % tasklist)
-        self.logger.info(' * Scenario              : %s' % getattr(self.scenario, 'name'))
-        if self.scenario.child_scenarios:
-            self.logger.info(' * Included Scenario(s)  : %s' % [getattr(sc, 'name') for sc in
-                                                                self.scenario.child_scenarios])
+        self.logger.info(' * Iterate Method        : %s' % self.scenario_graph.iterate_method)
+        for sc in self.scenario_graph:
+            self.logger.info(' * Scenario              : %s' % getattr(sc, 'name'))
         self.logger.info('-' * 79 + '\n')
 
     def _write_out_results(self):
-
-        # check if child scenario exist, update the new child_result and append to ch_result_list.
-        if self.scenario.child_scenarios:
-            child_result_list = []
-            for sc in self.scenario.child_scenarios:
-                ch_result_file_name = sc.name + '_results.yml'
-                ch_result_abs_name = os.path.join(self.data_folder, sc.name + '_results.yml')
-                file_mgmt('w', ch_result_abs_name, sc.profile())
-                # Adding child_result_list with results file in the RESULTS_FOLDER instead of absolute path
-                child_result_list.append(os.path.join(self.config['RESULTS_FOLDER'], ch_result_file_name))
-
-            # Add the child_result_list to the included_scenario_names property
-            self.scenario.__setattr__('included_scenario_names', child_result_list)
-
-        # Write the main scenario results
-        file_mgmt('w', self.results_file, self.scenario.profile())
+        """
+        write all results to sc.path_results.yml
+        if you are running sdf under data folder
+        or results folder, it will not generate
+        new sc.path_results.yml
+        """
+        sc: Scenario
+        for sc in self.scenario_graph:
+            # use filename because the scenario name could
+            # contain some special characters, which is not good for
+            # file generation
+            if preproc_path(self.config["RESULTS_FOLDER"]) in sc.fullpath or \
+                preproc_path(self.data_folder) in sc.fullpath \
+                    or sc.path == "results.yml":
+                ch_result_abs_name = os.path.join(self.data_folder, sc.path)
+            else:
+                ch_result_abs_name = os.path.join(self.data_folder, sc.path.split(".")[0] + '_results.yml')
+            file_mgmt('w', ch_result_abs_name, sc.profile())
+            # Adding child_result_list with results file in the RESULTS_FOLDER instead of absolute path
+        self._create_final_result_yml()
 
     def _print_footer(self, passed_tasks, failed_tasks, state):
 
@@ -600,14 +806,19 @@ class Teflo(LoggerMixin, TimeMixin):
                              failed_tasks)
         self.logger.info(' * Results Folder                 : %s' %
                          self.config['RESULTS_FOLDER'])
-
-        self.logger.info(' * Included Scenario Definition   : %s' % self.scenario.included_scenario_names)
-        self.logger.info(' * Final Scenario Definition      : %s' % os.path.join(self.config['RESULTS_FOLDER'],
-                                                                                 RESULTS_FILE))
+        self.logger.info(' * Iterate Method                 : %s' %
+                         self.scenario_graph.iterate_method)
+        for sc in self.scenario_graph:
+            self.logger.info(' * Scenario Definition            : %s' % sc.name)
+        self.logger.info(' * Final Scenario Definition      : %s' % self.final_results_yml_path)
         self.logger.info('-' * 79)
         self.logger.info('TEFLO RUN (RESULT=%s)' % state)
 
     def _archive_results(self):
+        """
+        This method copy all files under
+        data folder to results folder
+        """
 
         # archive everything from the data folder into the results folder
         os.system('cp -r %s/* %s' % (self.data_folder, self.config['RESULTS_FOLDER']))
@@ -617,4 +828,30 @@ class Teflo(LoggerMixin, TimeMixin):
         if self.static_inv_dir:
             if os.listdir(self.config['INVENTORY_FOLDER']):
                 inv_results_dir = os.path.join(self.config['RESULTS_FOLDER'], 'inventory')
-                os.system('cp -r %s/* %s' % (self.config['INVENTORY_FOLDER'], inv_results_dir))
+                if not os.path.samefile(self.config['INVENTORY_FOLDER'], inv_results_dir):
+                    os.system('cp -r %s/* %s' % (self.config['INVENTORY_FOLDER'], inv_results_dir))
+
+    def create_teflo_workspace(self, ctx, dirname):
+        """Clones the teflo examples git repo and copy the workspace files."""
+
+        # Download the teflo examples repo to cache folder
+        url = 'https://github.com/redhatqe/teflo_examples.git'
+        cmd = 'git clone ' + url + ' .teflo_cache'
+        result = exec_local_cmd(cmd)
+
+        if result[0] != 0:
+            click.echo("A problem occurred while cloning the teflo_examples project")
+            ctx.exit()
+
+        # Copy the files to the new directory and clean cache
+        copy_tree('./.teflo_cache/teflo_init/', dirname)
+        shutil.rmtree('.teflo_cache')
+
+    def showgraph(self, ctx, scenario_graph: ScenarioGraph):
+        """Show scenario graph includes structure"""
+        from termcolor import colored
+        click.echo(colored("Below is the structure of the Scenario Definition Files", "green"))
+        print('\n')
+        print('ITERATE METHOD:', scenario_graph.iterate_method)
+        print('\n\n\n')
+        print(colored(scenario_graph, "green"))
